import logging
from typing import List, Optional

import torch
import torch.nn.functional as F
from torch import nn

from ...constants import FEATURES, LABEL, LOGITS, NER_ANNOTATION, NER_TEXT, TOKEN_WORD_MAPPING, WORD_OFFSETS
from ..mlp import MLP
from ..utils import run_model
from .base import AbstractMultimodalFusionModel

logger = logging.getLogger(__name__)


class MultimodalFusionNER(AbstractMultimodalFusionModel):
    """
    Use MLP to fuse different models' features (single-modal and multimodal) for NER.
    Specifically, it adapts the features of each model to specified dimensions,
    concatenates the adapted features, and fuses the features through MLP.
    """

    def __init__(
        self,
        prefix: str,
        models: list,
        hidden_features: List[int],
        num_classes: int,
        adapt_in_features: str = "max",
        activation: Optional[str] = "gelu",
        dropout_prob: Optional[float] = 0.5,
        normalization: Optional[str] = "layer_norm",
        loss_weight: Optional[float] = None,
    ):
        """
        Parameters
        ----------
        prefix
            The fusion model's prefix
        models
            The individual models whose output features will be fused.
        hidden_features
            A list of integers representing the hidden feature dimensions. For example,
            [512, 128, 64] indicates three hidden MLP layers with their corresponding output
            feature dimensions.
        num_classes
            The number of classes.
        adapt_in_features
            Choice of how to adapt the features of each model. We now support
            - min
                Adapt all features to the minimum dimension. For example, if three models have
                feature dimensions [512, 768, 64], it will linearly map all the features to
                dimension 64.
            - max
                Adapt all features to the maximum dimension. For example, if three models have
                feature dimensions are [512, 768, 64], it will linearly map all the features to
                dimension 768.
        activation
            Name of activation function.
        dropout_prob
            Dropout probability.
        normalization
            Name of normalization function.
        loss_weight
            The weight of individual models.
        """
        super().__init__(
            prefix=prefix,
            models=models,
            loss_weight=loss_weight,
        )
        logger.debug("initializing MultimodalFusionNER")

        if loss_weight is not None:
            assert loss_weight > 0
        self.num_classes = num_classes

        self.ner_model = None
        self.tokenizer = None
        other_models = []
        for per_model in models:
            if per_model.prefix != NER_TEXT:
                other_models.append(per_model)
            else:
                self.ner_model = per_model
                self.tokenizer = per_model.tokenizer
        self.other_models = nn.ModuleList(other_models)
        raw_in_features = [per_model.out_features for per_model in models if per_model.prefix != NER_TEXT]
        if adapt_in_features is not None:
            if adapt_in_features == "min":
                base_in_feat = min(raw_in_features)
            elif adapt_in_features == "max":
                base_in_feat = max(raw_in_features)
            else:
                raise ValueError(f"unknown adapt_in_features: {adapt_in_features}")
            self.adapter = nn.ModuleList([nn.Linear(in_feat, base_in_feat) for in_feat in raw_in_features])
            in_features = base_in_feat * len(raw_in_features)
        else:
            self.adapter = nn.ModuleList([nn.Identity() for _ in range(len(raw_in_features))])
            in_features = sum(raw_in_features)
        assert len(self.adapter) == len(self.other_models)
        fusion_mlp = []
        for per_hidden_features in hidden_features:
            fusion_mlp.append(
                MLP(
                    in_features=in_features,
                    hidden_features=per_hidden_features,
                    out_features=per_hidden_features,
                    num_layers=1,
                    activation=activation,
                    dropout_prob=dropout_prob,
                    normalization=normalization,
                )
            )
            in_features = per_hidden_features
        self.fusion_mlp = nn.Sequential(*fusion_mlp)
        self.head = nn.Linear(in_features + self.ner_model.out_features, num_classes)
        self.out_features = in_features
        self.name_to_id = self.get_layer_ids()
        self.head_layer_names = [n for n, layer_id in self.name_to_id.items() if layer_id == 0]

    @property
    def label_key(self):
        return f"{NER_TEXT}_{LABEL}"

    def forward(
        self,
        batch: dict,
    ):
        """
        Parameters
        ----------
        batch
            A dictionary containing the input mini-batch data. The fusion model doesn't need to
            directly access the mini-batch data since it aims to fuse the individual models'
            output features.
        Returns
        -------
        It returns dictionary containing the fusion model's logits and features.
        """
        multimodal_features = []
        ner_output = run_model(self.ner_model, batch)
        for per_model, per_adapter in zip(self.other_models, self.adapter):
            per_output = run_model(per_model, batch)
            multimodal_features.append(per_adapter(per_output[per_model.prefix][FEATURES]))

        features = self.fusion_mlp(torch.cat(multimodal_features, dim=1))
        features = features.unsqueeze(dim=1).repeat(1, ner_output[self.ner_model.prefix][FEATURES].size()[1], 1)
        features = torch.cat((ner_output[self.ner_model.prefix][FEATURES], features), dim=-1)

        logits = self.head(features)
        logits_label = torch.argmax(F.log_softmax(logits, dim=-1), dim=-1)
        fusion_output = {
            self.prefix: {
                LOGITS: logits,
                FEATURES: features,
                NER_ANNOTATION: logits_label,
                TOKEN_WORD_MAPPING: ner_output[self.ner_model.prefix][TOKEN_WORD_MAPPING],
                WORD_OFFSETS: ner_output[self.ner_model.prefix][WORD_OFFSETS],
            }
        }

        return fusion_output



from .base import BaseLearner
from .ensemble import EnsembleLearner
from .few_shot_svm import FewShotSVMLearner
from .matching import MatchingLearner
from .ner import NERLearner
from .object_detection import ObjectDetectionLearner
from .semantic_segmentation import SemanticSegmentationLearner



_base_ = "./yolox_tiny_8xb8-300e_coco.py"

# model settings
model = dict(
    backbone=dict(deepen_factor=0.33, widen_factor=0.25, use_depthwise=True),
    neck=dict(in_channels=[64, 128, 256], out_channels=64, num_csp_blocks=1, use_depthwise=True),
    bbox_head=dict(in_channels=64, feat_channels=64, use_depthwise=True),
)



import logging
from typing import Callable, Dict, List, Optional, Union

import lightning.pytorch as pl
import torch
import torch.nn.functional as F
import torchmetrics
from lightning.pytorch.strategies import DeepSpeedStrategy
from lightning.pytorch.utilities import grad_norm
from torch import nn
from torch.nn.modules.loss import _Loss
from torchmetrics.aggregation import BaseAggregator

from ..constants import (
    AUG_LOGITS,
    LM_TARGET,
    LOGITS,
    MULTIMODAL_FEATURES,
    MULTIMODAL_FEATURES_POST_AUG,
    MULTIMODAL_FEATURES_PRE_AUG,
    ORI_LOGITS,
    T_FEW,
    TEMPLATE_LOGITS,
    VAE_MEAN,
    VAE_VAR,
    WEIGHT,
)
from ..data.mixup import MixupModule, multimodel_mixup
from ..models.utils import run_model
from .lr import apply_layerwise_lr_decay, apply_single_lr, apply_two_stages_lr, get_lr_scheduler
from .metrics import Coverage
from .utils import get_optimizer

logger = logging.getLogger(__name__)


class LitModule(pl.LightningModule):
    """
    Control the loops for training, evaluation, and prediction. This module is independent of
    the model definition. This class inherits from the Pytorch Lightning's LightningModule:
    https://lightning.ai/docs/pytorch/stable/common/lightning_module.html
    """

    def __init__(
        self,
        model: nn.Module,
        optim_type: Optional[str] = None,
        lr_choice: Optional[str] = None,
        lr_schedule: Optional[str] = None,
        lr: Optional[float] = None,
        lr_decay: Optional[float] = None,
        end_lr: Optional[Union[float, int]] = None,
        lr_mult: Optional[Union[float, int]] = None,
        weight_decay: Optional[float] = None,
        warmup_steps: Optional[int] = None,
        loss_func: Optional[_Loss] = None,
        validation_metric: Optional[torchmetrics.Metric] = None,
        validation_metric_name: Optional[str] = None,
        custom_metric_func: Callable = None,
        test_metric: Optional[torchmetrics.Metric] = None,
        peft: Optional[str] = None,
        trainable_param_names: Optional[List] = None,
        mixup_fn: Optional[MixupModule] = None,
        mixup_off_epoch: Optional[int] = 0,
        model_postprocess_fn: Callable = None,
        skip_final_val: Optional[bool] = False,
        track_grad_norm: Optional[Union[int, str]] = -1,
        cross_modal_align: Optional[str] = None,
        cross_modal_align_weight: Optional[float] = 0,
        automatic_optimization: Optional[bool] = True,
        accumulate_grad_batches: Optional[int] = None,
        gradient_clip_val: Optional[float] = None,
        gradient_clip_algorithm: Optional[str] = None,
        use_aug_optim: Optional[bool] = False,
        aug_loss_func: Optional[_Loss] = None,
        aug_lr: Optional[float] = None,
        aug_weight_decay: Optional[float] = None,
        aug_optim_type: Optional[str] = None,
    ):
        """
        Parameters
        ----------
        model
            A Pytorch model
        optim_type
            Optimizer type. We now support:
            - adamw
            - adam
            - sgd
        lr_choice
            How to set each layer's learning rate. If not specified, the default is a single
            learnng rate for all layers. Otherwise, we now support two choices:
            - two_stages
                The layers in the pretrained models have a small learning rate (lr * lr_mult),
                while the newly added head layers use the provided learning rate.
            - layerwise_decay
                The layers have decreasing learning rate from the output end to the input end.
                The intuition is that later layers are more task-related, hence larger learning rates.
        lr_schedule
            Learning rate schedule. We now support:
            - cosine_decay
                Linear warmup followed by cosine decay
            - polynomial_decay
                Linear warmup followed by polynomial decay
        lr
            Learning rate.
        lr_decay
            The learning rate decay factor (0, 1). It is used only when lr_choice is "layerwise_decay".
        end_lr
            The final learning rate after decay.
        lr_mult
            The learning rate multiplier (0, 1). It is used only when lr_choice is "two_stages".
        weight_decay
            The weight decay to regularize layer weights' l2 norm.
        warmup_steps
            How many steps to warmup learning rate. If a float (0, 1), it would represent the
            percentage of steps over all the training steps. The actual number is calculated as
            "int(warmup_steps * max_steps)". If an integer, it would be the exact step number.
        loss_func
            A Pytorch loss module, e.g., nn.CrossEntropyLoss().
        validation_metric
            A torchmetrics module used in the validation stage, e.g., torchmetrics.Accuracy().
        validation_metric_name
            Name of validation metric in case that validation_metric is a aggregation metric,
            e.g., torchmetrics.MeanMetric, whose name can't reflect the real metric name.
        custom_metric_func
            A customized metric function in case that torchmetrics doesn't have the metric.
            It is generally used together with torchmetrics' aggregators, e.g., torchmetrics.MeanMetric.
            Refer to https://github.com/PyTorchLightning/metrics/blob/master/torchmetrics/aggregation.py
        test_metric
            A torchmetrics module used in the test stage, e.g., torchmetrics.Accuracy().
        peft
            Whether to use efficient finetuning strategies. This will be helpful for fast finetuning of large backbones.
            We support options such as:

            - bit_fit (only finetune the bias terms)
            - norm_fit (only finetune the weights in norm layers / bias layer)
            - lora, lora_bias, lora_norm (only finetunes decomposition matrices inserted into model, in combination with either bit_fit or norm_fit)
            - ia3, ia3_bias, ia3_norm (adds vector that scales activations by learned vectors, in combination with either bit_fit or norm_fit)
            - None (do not use efficient finetuning strategies)
        track_grad_norm
            Track the p-norm of gradients during training. May be set to 'inf'  infinity-norm.
            If using Automatic Mixed Precision (AMP), the gradients will be unscaled before logging them.

        """
        super().__init__()
        self.save_hyperparameters(
            ignore=[
                "model",
                "validation_metric",
                "test_metric",
                "loss_func",
                "model_postprocess_fn",
                "mixup_fn",
                "trainable_param_names",
                "custom_metric_func",
                "aug_loss_func",
            ]
        )
        self.model = model
        self.validation_metric = validation_metric
        self.validation_metric_name = f"val_{validation_metric_name}"
        self.loss_func = loss_func
        self.mixup_fn = mixup_fn
        if isinstance(validation_metric, BaseAggregator) and custom_metric_func is None:
            raise ValueError(
                f"validation_metric {validation_metric} is an aggregation metric,"
                "which must be used with a customized metric function."
            )
        self.custom_metric_func = custom_metric_func
        self.model_postprocess_fn = model_postprocess_fn
        self.trainable_param_names = trainable_param_names if trainable_param_names else []
        self.skip_final_val = skip_final_val
        self.automatic_optimization = automatic_optimization
        self.aug_loss_func = aug_loss_func
        if self.hparams.cross_modal_align:
            assert self.hparams.cross_modal_align_weight > 0
            logger.debug(f"Cross modal alignment mode: {self.hparams.cross_modal_align}")
            logger.debug(f"Cross modal alignment loss weight: {self.hparams.cross_modal_align_weight}")

    def _compute_template_loss(
        self,
        per_output: Dict,
        label: torch.Tensor,
    ):
        logits = per_output[TEMPLATE_LOGITS]
        choices_scores = per_output[LOGITS]
        lm_target = per_output[LM_TARGET]

        bs = lm_target.size(0)
        num_choices = lm_target.size(1)

        lm_loss = F.cross_entropy(
            logits[range(bs), label].flatten(0, 1),
            lm_target[range(bs), label].flatten(0, 1),
        )
        if self.model.mc_loss > 0:
            mc_loss = F.cross_entropy(choices_scores, label)
        else:
            mc_loss = 0.0

        if self.model.unlikely_loss > 0:
            cand_loglikely = -F.cross_entropy(logits.flatten(0, 2), lm_target.flatten(0, 2), reduction="none").view(
                bs, num_choices, -1
            )
            cand_loglikely += (lm_target < 0) * -100
            cand_loglikely[range(bs), label] = -100
            unlikely_loss = -torch.log(1 - torch.exp(cand_loglikely) + 1e-2).sum() / (cand_loglikely != -100).sum()
        else:
            unlikely_loss = 0.0

        return lm_loss + mc_loss * self.model.mc_loss + unlikely_loss * self.model.unlikely_loss

    def _compute_cross_modal_align_loss(self, multimodal_features):
        if self.hparams.cross_modal_align == "positive_only":
            kl_loss = nn.KLDivLoss(reduction="batchmean", log_target=True)
            loss = 0
            num = 0
            for i in range(len(multimodal_features)):
                # input should be a distribution in the log space
                a = F.log_softmax(multimodal_features[i], dim=1)
                # kl divergence is not symmetric, so need to compute both (i, j) and (j, i)
                for j in range(len(multimodal_features)):
                    if i == j:
                        continue
                    # input should be a distribution in the log space
                    b = F.log_softmax(multimodal_features[j], dim=1)
                    loss += kl_loss(a, b)
                    num += 1
            return self.hparams.cross_modal_align_weight * loss / num
        else:
            raise ValueError(f"Unsupported cross modal alignment loss: {self.hparams.cross_modal_align}.")

    def _compute_loss(
        self,
        output: Dict,
        label: torch.Tensor,
    ):
        loss = 0
        for per_prefix, per_output in output.items():
            weight = per_output[WEIGHT] if WEIGHT in per_output else 1
            if (
                TEMPLATE_LOGITS in per_output and self.model.prefix == T_FEW
            ):  # Do only add template loss if T-Few. #TODO Add compatibility to Fusion models.
                loss += self._compute_template_loss(per_output, label) * weight
            else:
                if self.training and self.hparams.use_aug_optim and per_prefix.startswith("fusion"):
                    label = label.tile((2,))
                loss += (
                    self.loss_func(
                        input=per_output[LOGITS].squeeze(dim=1),
                        target=label,
                    )
                    * weight
                )

        if self.hparams.cross_modal_align:
            loss += self._compute_cross_modal_align_loss(
                multimodal_features=output[self.model.prefix][MULTIMODAL_FEATURES]
            )

        if self.training and self.hparams.use_aug_optim:
            loss += self.aug_loss_func(
                pre_aug=output[self.model.prefix][MULTIMODAL_FEATURES_PRE_AUG],
                post_aug=output[self.model.prefix][MULTIMODAL_FEATURES_POST_AUG],
                vae_mean=output[self.model.prefix][VAE_MEAN],
                vae_var=output[self.model.prefix][VAE_VAR],
                ori_logits=output[self.model.prefix][ORI_LOGITS],
                aug_logits=output[self.model.prefix][AUG_LOGITS],
            )

        return loss

    def _compute_metric_score(
        self,
        metric: torchmetrics.Metric,
        custom_metric_func: Callable,
        logits: torch.Tensor,
        label: torch.Tensor,
    ):
        if isinstance(
            metric,
            (
                torchmetrics.classification.BinaryAUROC,
                torchmetrics.classification.BinaryAveragePrecision,
                torchmetrics.classification.BinaryF1Score,
                Coverage,
            ),
        ):
            prob = F.softmax(logits.float(), dim=1)
            metric.update(preds=prob[:, 1], target=label)  # only for binary classification
        elif isinstance(metric, BaseAggregator):
            metric.update(custom_metric_func(logits, label))
        else:
            metric.update(logits.squeeze(dim=1).float(), label)

    def _shared_step(
        self,
        batch: Dict,
    ):
        label = batch[self.model.label_key]
        if self.mixup_fn is not None:
            self.mixup_fn.mixup_enabled = self.training & (self.current_epoch < self.hparams.mixup_off_epoch)
            batch, label = multimodel_mixup(batch=batch, model=self.model, mixup_fn=self.mixup_fn)
        output = run_model(self.model, batch)
        loss = self._compute_loss(output=output, label=label)
        return output, loss

    def training_step(self, batch, batch_idx):
        """
        Per training step. This function is registered by LightningModule.
        Refer to https://lightning.ai/docs/pytorch/stable/common/lightning_module.html#training-loop

        Parameters
        ----------
        batch
            A dictionary containing the mini-batch data, including both input data and
            ground-truth labels. The mini-batch data are passed to each individual model,
            which indexes its required input data by keys with its model prefix. The
            ground-truth labels are used here to compute the training loss.
        batch_idx
            Index of mini-batch.

        Returns
        -------
        Average loss of the mini-batch data.
        """
        output, loss = self._shared_step(batch)

        if not self.automatic_optimization:
            if self.hparams.use_aug_optim:
                optimizer, aug_optimizer = self.optimizers()
            else:
                optimizer = self.optimizers()
                aug_optimizer = None

            lr_scheduler = self.lr_schedulers()
            loss = loss / self.hparams.accumulate_grad_batches
            self.manual_backward(loss)

            if (batch_idx + 1) % self.hparams.accumulate_grad_batches == 0 or self.trainer.is_last_batch:
                optimizer.step()
                optimizer.zero_grad()
                lr_scheduler.step()

                if aug_optimizer is not None:
                    aug_optimizer.step()
                    aug_optimizer.zero_grad()

        self.log("train_loss", loss)
        return loss

    def on_validation_start(self) -> None:
        if self.skip_final_val and self.trainer.should_stop:
            self.log(
                self.validation_metric_name,
                self.validation_metric,
                on_step=False,
                on_epoch=True,
            )
            return None
        else:
            return super().on_validation_start()

    def validation_step(self, batch, batch_idx):
        """
        Per validation step. This function is registered by LightningModule.
        Refer to https://lightning.ai/docs/pytorch/stable/common/lightning_module.html#validation-loop

        Parameters
        ----------
        batch
            A dictionary containing the mini-batch data, including both input data and
            ground-truth labels. The mini-batch data are passed to each individual model,
            which indexes its required input data by keys with its model prefix. The
            ground-truth labels are used here to compute the validation loss and metric.
            The validation metric is used for top k model selection and early stopping.
        batch_idx
            Index of mini-batch.
        """
        output, loss = self._shared_step(batch)
        if self.model_postprocess_fn:
            output = self.model_postprocess_fn(output)
        # By default, on_step=False and on_epoch=True
        self.log("val_loss", loss)
        self._compute_metric_score(
            metric=self.validation_metric,
            custom_metric_func=self.custom_metric_func,
            logits=output[self.model.prefix][LOGITS],
            label=batch[self.model.label_key],
        )
        self.log(
            self.validation_metric_name,
            self.validation_metric,
            on_step=False,
            on_epoch=True,
        )

    def predict_step(self, batch, batch_idx, dataloader_idx=0):
        """
        Per prediction step. This function is registered by LightningModule.
        Refer to https://lightning.ai/docs/pytorch/stable/common/lightning_module.html#prediction-loop

        Parameters
        ----------
        batch
            A dictionary containing the mini-batch data.
            The mini-batch data are passed to each individual model,
            which indexes its required input data by keys with its model prefix.
            Ground-truth labels are not needed for prediction.
        batch_idx
            Index of mini-batch.
        dataloader_idx
            Index of dataloader.
        Returns
        -------
        A dictionary with the mini-batch's logits and features.
        """
        output = run_model(self.model, batch)
        if self.model_postprocess_fn:
            output = self.model_postprocess_fn(output)

        return output[self.model.prefix]

    def configure_optimizers(self):
        """
        Configure optimizer. This function is registered by LightningModule.
        Refer to https://lightning.ai/docs/pytorch/stable/common/lightning_module.html#configure-optimizers
        Returns
        -------
        [optimizer]
            Optimizer.
        [sched]
            Learning rate scheduler.
        """
        kwargs = dict(
            model=self.model,
            lr=self.hparams.lr,
            weight_decay=self.hparams.weight_decay,
            exclude_keys=[
                "augmenter"
            ],  # exclude augmenter parameters from the optimizer as they would use an independent optimizer
        )
        if self.hparams.lr_choice == "two_stages":
            logger.debug("applying 2-stage learning rate...")
            grouped_parameters = apply_two_stages_lr(
                lr_mult=self.hparams.lr_mult,
                return_params=True,
                **kwargs,
            )
        elif self.hparams.lr_choice == "layerwise_decay":
            logger.debug("applying layerwise learning rate decay...")
            grouped_parameters = apply_layerwise_lr_decay(
                lr_decay=self.hparams.lr_decay,
                peft=self.hparams.peft,
                trainable_param_names=self.trainable_param_names,
                **kwargs,
            )
        else:
            logger.debug("applying single learning rate...")
            grouped_parameters = apply_single_lr(
                peft=self.hparams.peft,
                trainable_param_names=self.trainable_param_names,
                **kwargs,
            )

        optimizer = get_optimizer(
            optim_type=self.hparams.optim_type,
            optimizer_grouped_parameters=grouped_parameters,
            lr=self.hparams.lr,
            weight_decay=self.hparams.weight_decay,
        )

        logger.debug(f"trainer.max_steps: {self.trainer.max_steps}")
        if self.trainer.max_steps is None or -1:
            if isinstance(self.trainer.strategy, DeepSpeedStrategy):
                max_steps = 1
            else:
                accumulate_grad_batches = (
                    self.trainer.accumulate_grad_batches
                    if self.automatic_optimization
                    else self.hparams.accumulate_grad_batches
                )
                max_steps = (
                    len(self.trainer.datamodule.train_dataloader())
                    * self.trainer.max_epochs
                    // accumulate_grad_batches
                    // self.trainer.num_devices  # Account for distributed training
                )
                logger.debug(
                    f"len(trainer.datamodule.train_dataloader()): {len(self.trainer.datamodule.train_dataloader())}"
                )
                logger.debug(f"trainer.max_epochs: {self.trainer.max_epochs}")
                logger.debug(f"accumulate_grad_batches: {accumulate_grad_batches}")
                logger.debug(f"num_devices: {self.trainer.num_devices}")
        else:
            max_steps = self.trainer.max_steps

        logger.debug(f"max steps: {max_steps}")

        warmup_steps = self.hparams.warmup_steps
        if isinstance(warmup_steps, float):
            warmup_steps = int(max_steps * warmup_steps)

        logger.debug(f"warmup steps: {warmup_steps}")
        logger.debug(f"lr_schedule: {self.hparams.lr_schedule}")
        scheduler = get_lr_scheduler(
            optimizer=optimizer,
            num_max_steps=max_steps,
            num_warmup_steps=warmup_steps,
            lr_schedule=self.hparams.lr_schedule,
            end_lr=self.hparams.end_lr,
        )

        sched = {"scheduler": scheduler, "interval": "step"}
        ret_optimizers = [optimizer]
        ret_schedulers = [sched]
        if self.hparams.use_aug_optim:
            logger.debug("initializing augment optimizer")
            # augmenter's optimizer
            aug_grouped_parameters = apply_single_lr(
                model=self.model.augmenter,
                lr=self.hparams.aug_lr,
                weight_decay=self.hparams.aug_weight_decay,
            )
            aug_optimizer = get_optimizer(
                optim_type=self.hparams.aug_optim_type,
                optimizer_grouped_parameters=aug_grouped_parameters,
                lr=self.hparams.aug_lr,
                weight_decay=self.hparams.aug_weight_decay,
            )
            ret_optimizers.append(aug_optimizer)

        logger.debug("done configuring optimizer and scheduler")
        return ret_optimizers, ret_schedulers

    def on_before_optimizer_step(self, optimizer):
        # If using mixed precision, the gradients are already unscaled here
        # TODO: apply gradient clip only to the target optimizer
        if not self.automatic_optimization and self.hparams.gradient_clip_val > 0:
            self.clip_gradients(
                optimizer,
                gradient_clip_val=self.hparams.gradient_clip_val,
                gradient_clip_algorithm=self.hparams.gradient_clip_algorithm,
            )
        if self.hparams.track_grad_norm != -1:
            self.log_dict(grad_norm(self, norm_type=self.hparams.track_grad_norm))



import logging
import random
from typing import Any, Dict, List, Optional, Union

import numpy as np
from torch import nn

from ..constants import COLUMN, NUMERICAL
from .collator import StackCollator

logger = logging.getLogger(__name__)


class NumericalProcessor:
    """
    Prepare numerical data for the model specified by "prefix".
    For multiple models requiring numerical data, we need to create a NumericalProcessor
    for each related model so that they will have independent input.
    """

    def __init__(
        self,
        model: nn.Module,
        merge: Optional[str] = "concat",
        requires_column_info: bool = False,
        dropout: Optional[float] = 0,
    ):
        """
        Parameters
        ----------
        model
            The model for which this processor would be created.
        merge
            How to merge numerical features from multiple columns in a multimodal pd.DataFrame.
            Currently, it only supports one choice:
            - concat
                Concatenate the numerical features.
        requires_column_info
            Whether to require feature column information in dataloader.
        """
        logger.debug(f"initializing numerical processor for model {model.prefix}")
        self.prefix = model.prefix
        self.merge = merge
        self.requires_column_info = requires_column_info
        self.numerical_fill_values = model.numerical_fill_values
        self.dropout = dropout
        assert 0 <= self.dropout <= 1
        if self.dropout > 0:
            logger.debug(f"numerical value dropout probability: {self.dropout}")
            logger.debug(f"dropped values will be replaced by {self.numerical_fill_values}")

    @property
    def numerical_key(self):
        return f"{self.prefix}_{NUMERICAL}"

    @property
    def numerical_column_prefix(self):
        return f"{self.numerical_key}_{COLUMN}"

    def collate_fn(self, numerical_column_names: List) -> Dict:
        """
        Collate individual samples into a batch. Here it stacks numerical features.
        This function will be used when creating Pytorch DataLoader.

        Returns
        -------
        A dictionary containing one model's collator function for numerical features.
        """
        fn = {}
        if self.requires_column_info:
            assert numerical_column_names, "Empty numerical column names."
            for col_name in numerical_column_names:
                fn[f"{self.numerical_column_prefix}_{col_name}"] = StackCollator()

        fn[self.numerical_key] = StackCollator()

        return fn

    def process_one_sample(
        self,
        numerical_features: Dict[str, float],
        is_training: bool,
    ) -> Dict:
        """
        Process one sample's numerical features.
        Here it converts numerical features to a NumPy array.

        Parameters
        ----------
        numerical_features
            Numerical features of one sample.
        is_training
            Whether to do processing in the training mode.

        Returns
        -------
        A dictionary containing the processed numerical features.
        """
        ret = {}
        if self.requires_column_info:
            # TODO: consider moving this for loop into __init__() since each sample has the same information.
            for i, col_name in enumerate(numerical_features.keys()):
                ret[f"{self.numerical_column_prefix}_{col_name}"] = i

        if is_training and self.dropout > 0:
            numerical_features_copy = dict()
            for k, v in numerical_features.items():
                if random.uniform(0, 1) <= self.dropout:
                    numerical_features_copy[k] = self.numerical_fill_values[k]
                else:
                    numerical_features_copy[k] = v
            numerical_features = numerical_features_copy

        if self.merge == "concat":
            ret[self.numerical_key] = np.array(list(numerical_features.values()), dtype=np.float32)
        else:
            raise ValueError(f"Unknown merging type: {self.merge}")

        return ret

    def __call__(
        self,
        numerical_features: Dict[str, float],
        sub_dtypes: Dict[str, str],
        is_training: bool,
    ) -> Dict:
        """
        Extract one sample's numerical features and customize it for a specific model.

        Parameters
        ----------
        numerical_features
            Numerical features of one sample.
        sub_dtypes
            The sub data types of all numerical columns.
        is_training
            Whether to do processing in the training mode.

        Returns
        -------
        A dictionary containing one sample's processed numerical features.
        """
        return self.process_one_sample(
            numerical_features=numerical_features,
            is_training=is_training,
        )



import argparse
import os
import shutil

import requests


def get_root_dir(output_dir=None, new_folder_name=None):
    if not output_dir:
        root_dir = "./"
    elif os.path.isdir(output_dir):
        if new_folder_name:
            root_dir = os.path.join(output_dir, new_folder_name)
            if not os.path.exists(root_dir):
                os.mkdir(root_dir)
        else:
            root_dir = output_dir
    else:
        raise ValueError(f"{output_dir} is not a valid directory")
    return root_dir


def get_fname_from_path_or_url(path_or_url):
    fname_start = path_or_url.rfind("/") + 1
    fname = path_or_url[fname_start:]
    return fname


def download_one_url(url, root_dir, fname=None):
    if not fname:
        fname = get_fname_from_path_or_url(url)
    output_path = os.path.join(root_dir, fname)
    print(f"Downloading {fname}...")

    r = requests.get(url, timeout=(10, 1000))
    with open(output_path, "wb") as f:
        f.write(r.content)

    return output_path


def download_urls(urls, root_dir, fnames=[]):
    output_paths = []
    if isinstance(urls, str):
        urls = [urls]
    for i, url in enumerate(urls):
        output_path = download_one_url(url, root_dir, fname=fnames[i] if fnames else None)
        output_paths.append(output_path)
    return output_paths


def unpack(archived_file_paths, root_dir):
    for archived_file_path in archived_file_paths:
        fname = get_fname_from_path_or_url(archived_file_path)
        print(f"extracting {fname}...")
        shutil.unpack_archive(archived_file_path, root_dir)


def remove_archived_file_paths(archived_file_paths):
    for archived_file_path in archived_file_paths:
        fname = get_fname_from_path_or_url(archived_file_path)
        print(f"removing {fname}...")
        os.remove(archived_file_path)


def prepare_dataset(output_dir, new_folder_name, urls, fnames=[]):
    root_dir = get_root_dir(output_dir=output_dir, new_folder_name=new_folder_name)
    archived_file_paths = download_urls(urls, root_dir, fnames=fnames)
    unpack(archived_file_paths, root_dir)
    remove_archived_file_paths(archived_file_paths)


def prepare_coco17(output_dir):
    print("Preparing COCO17 dataset...")
    urls = [
        "http://images.cocodataset.org/zips/train2017.zip",
        "http://images.cocodataset.org/zips/val2017.zip",
        "http://images.cocodataset.org/zips/test2017.zip",
        "http://images.cocodataset.org/zips/unlabeled2017.zip",
        "http://images.cocodataset.org/annotations/annotations_trainval2017.zip",
        "http://images.cocodataset.org/annotations/stuff_annotations_trainval2017.zip",
        "http://images.cocodataset.org/annotations/image_info_test2017.zip",
        "http://images.cocodataset.org/annotations/image_info_unlabeled2017.zip",
    ]
    prepare_dataset(output_dir=output_dir, new_folder_name="coco17", urls=urls)


def prepare_voc07(output_dir):
    print("Preparing VOC07 dataset...")
    urls = [
        "http://host.robots.ox.ac.uk/pascal/VOC/voc2007/VOCtrainval_06-Nov-2007.tar",
        "http://host.robots.ox.ac.uk/pascal/VOC/voc2007/VOCtest_06-Nov-2007.tar",
    ]
    prepare_dataset(output_dir=output_dir, new_folder_name=None, urls=urls)


def prepare_voc12(output_dir):
    print("Preparing VOC12 dataset...")
    urls = [
        "http://host.robots.ox.ac.uk/pascal/VOC/voc2012/VOCtrainval_11-May-2012.tar",
    ]
    prepare_dataset(output_dir=output_dir, new_folder_name=None, urls=urls)


def prepare_voc0712(output_dir):
    prepare_voc07(output_dir)
    prepare_voc12(output_dir)


def prepare_watercolor(output_dir):
    print("Preparing Watercolor dataset...")
    urls = [
        "http://www.hal.t.u-tokyo.ac.jp/~inoue/projects/cross_domain_detection/datasets/watercolor.zip",
    ]
    prepare_dataset(output_dir=output_dir, new_folder_name=None, urls=urls)


def prepare_pothole(output_dir):
    print("Preparing Pothole dataset...")
    # urls = ["https://www.kaggle.com/datasets/andrewmvd/pothole-detection/download?datasetVersionNumber=1"]
    urls = ["https://automl-mm-bench.s3.amazonaws.com/object_detection/dataset/pothole.zip"]
    prepare_dataset(output_dir=output_dir, new_folder_name=None, urls=urls, fnames=[])


def main(dataset_name, output_dir):
    if dataset_name.lower() in ["coco", "coco17", "coco2017"]:
        prepare_coco17(output_dir=output_dir)
    elif dataset_name.lower() in ["voc", "voc0712"]:
        prepare_voc0712(output_dir=output_dir)
    elif dataset_name.lower() in ["voc07", "voc2007"]:
        prepare_voc07(output_dir=output_dir)
    elif dataset_name.lower() in ["voc12", "voc2012"]:
        prepare_voc12(output_dir=output_dir)
    elif dataset_name.lower() in ["watercolor"]:
        prepare_watercolor(output_dir=output_dir)
    elif dataset_name.lower() in ["pothole"]:
        prepare_pothole(output_dir=output_dir)
    else:
        raise ValueError(f"This dataset name is not supported: {dataset_name}")


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("-d", "--dataset_name", type=str)
    parser.add_argument("-o", "--output_dir", default="./", type=str)
    args = parser.parse_args()

    main(args.dataset_name, args.output_dir)



import logging
import os
import warnings
from typing import Any, Callable, Dict, List, Optional, Union

import numpy as np
import PIL
import pytesseract
from numpy.typing import NDArray
from torch import nn
from torchvision import transforms

from ..constants import BBOX, DOCUMENT_PDF
from ..models.utils import get_pretrained_tokenizer
from .collator import PadCollator
from .process_image import ImageProcessor

logger = logging.getLogger(__name__)


class DocumentProcessor(ImageProcessor):
    """
    Prepare document data for Document Classification.
    OCR (Optical character recognition) is applied to get the document texts and bounding boxes.
    Both texts and images will be processed.
    """

    def __init__(
        self,
        model: nn.Module,
        train_transforms: Union[List[str], Callable, List[Callable]],
        val_transforms: Union[List[str], Callable, List[Callable]],
        size: Optional[int] = None,
        text_max_len: Optional[int] = 512,
        missing_value_strategy: Optional[str] = "zero",
    ):
        """
        Parameters
        ----------
        model
            The model using this data processor.
        train_transforms
            A list of image transforms used in training. Note that the transform order matters.
        val_transforms
            A list of image transforms used in validation/test/prediction. Note that the transform order matters.
        size
            The width / height of a square image.
        text_max_len
            The max text length of tokenizer. Default: 512.
        missing_value_strategy
            How to deal with a missing document. We now support:
            - skip
                Skip this sample
            -zero
                Use a document image with zero pixels.
        """
        self.prefix = model.prefix
        self.text_token_ids_key = model.text_token_ids_key
        self.text_attention_mask_key = model.text_attention_mask_key
        self.text_bbox_key = model.text_bbox_key
        self.text_segment_ids_key = model.text_segment_ids_key
        self.document_pixel_value_key = model.document_pixel_value_key

        # For document image processing.
        self.size = size
        self.train_transforms = train_transforms
        self.val_transforms = val_transforms
        self.mean = model.image_mean
        self.std = model.image_std
        self.normalization = transforms.Normalize(self.mean, self.std)
        self.train_processor = self.construct_image_processor(
            size=self.size, normalization=self.normalization, image_transforms=self.train_transforms
        )
        self.val_processor = self.construct_image_processor(
            size=self.size, normalization=self.normalization, image_transforms=self.val_transforms
        )

        self.missing_value_strategy = missing_value_strategy

        # Store OCR results
        self.documents = {}

        # Whether only text is used (automatically detected).
        # If True, normal text foundation models (e.g., bert-base) can be used.
        self.is_text_only_flag = model.is_text_only_flag

        self.pad_token_box = [0, 0, 0, 0]  # cls box token
        self.sep_token_box = [1000, 1000, 1000, 1000]  # sep box token

        # For document text processing.
        # Disable tokenizer parallelism
        os.environ["TOKENIZERS_PARALLELISM"] = "false"
        self.tokenizer_name = model.tokenizer_name
        self.tokenizer = model.tokenizer
        if text_max_len is None or text_max_len <= 0:
            self.text_max_len = self.tokenizer.model_max_length
        else:
            if text_max_len < self.tokenizer.model_max_length:
                warnings.warn(
                    f"provided max length: {text_max_len} "
                    f"is smaller than {model.checkpoint_name}'s default: {self.tokenizer.model_max_length}"
                )
            self.text_max_len = min(text_max_len, self.tokenizer.model_max_length)

        self.tokenizer.model_max_length = self.text_max_len

    def collate_fn(self, text_column_names: Optional[List] = None) -> Dict:
        """
        Collate multimodal features into a batch.
        This function will be used when creating Pytorch DataLoader.

        Returns
        -------
        A dictionary containing one model's collator function for multimodal data.
        """
        fn = {}

        fn.update(
            {
                self.text_token_ids_key: PadCollator(pad_val=self.tokenizer.pad_token_id),
                self.text_attention_mask_key: PadCollator(pad_val=0),
                self.text_bbox_key: PadCollator(pad_val=0),
                self.text_segment_ids_key: PadCollator(pad_val=0),
            }
        )
        # If not text only, document images will be used.
        if not self.is_text_only_flag:
            fn.update(
                {
                    self.document_pixel_value_key: PadCollator(pad_val=0),
                }
            )
        return fn

    @staticmethod
    def normalize_box(box, width, height):
        """
        Normalize the bounding boxes.

        Parameters
        ----------
        box
            The bounding box to be processed.
        width
            Width of the document image.
        height
            Height of the document image.

        Returns
        -------
        A normalized bounding box.
        """
        return [
            int(1000 * (box[0] / width)),
            int(1000 * (box[1] / height)),
            int(1000 * (box[2] / width)),
            int(1000 * (box[3] / height)),
        ]

    def apply_ocr(self, doc_image):
        """
        Apply OCR on document images to ecognize and “read” the text embedded in document images.
        Specifically, Python-tesseract, a wrapper for Google's Tesseract-OCR Engine, will be used.

        Parameters
        ----------
        doc_image
            The document image.

        Returns
        -------
        A dictionary with recognized text and corresponding bounding boxes.
        """
        results = {}
        width, height = doc_image.size
        # apply ocr to the document image.
        ocr_df = pytesseract.image_to_data(doc_image, output_type="data.frame")
        float_cols = ocr_df.select_dtypes("float").columns
        ocr_df = ocr_df.dropna().reset_index(drop=True)
        ocr_df[float_cols] = ocr_df[float_cols].round(0).astype(int)
        ocr_df = ocr_df.replace(r"^\s*$", np.nan, regex=True)
        ocr_df = ocr_df.dropna().reset_index(drop=True)

        # get the words and actual (unnormalized) bounding boxes.
        words = list(ocr_df.text)
        words = [str(w) for w in words]
        coordinates = ocr_df[["left", "top", "width", "height"]]
        actual_boxes = []
        for idx, row in coordinates.iterrows():
            # the row comes in (left, top, width, height) format.
            x, y, w, h = tuple(row)
            # we turn it into (left, top, left+width, top+height) to get the actual box.
            actual_box = [x, y, x + w, y + h]
            actual_boxes.append(actual_box)

        # normalize the bounding boxes.
        boxes = []
        for box in actual_boxes:
            boxes.append(self.normalize_box(box, width, height))

        assert len(words) == len(boxes)

        results["words"] = words
        results[BBOX] = boxes
        return results

    def get_ocr_features(self, doc_path, doc_image):
        """
        Get the OCR features, and avoid running OCR multiple times.

        Parameters
        ----------
        doc_path
            The document path.
        doc_image
            The document image.

        Returns
        -------
        Extracted words, image pixel features, and bounding boxes.
        """
        # The OCR process is time-consuming, so apply OCR on each image only once.
        if doc_path not in self.documents:
            ocr_res = self.apply_ocr(doc_image)
            # store the ocr results.
            self.documents.update({doc_path: ocr_res})
        else:
            # reuse the ocr results.
            ocr_res = self.documents[doc_path]

        words = ocr_res["words"]
        normalized_word_boxes = ocr_res[BBOX]

        return words, normalized_word_boxes

    def process_one_sample(
        self,
        document_features: Dict[str, Union[NDArray, list]],
        feature_modalities: Dict[str, Union[NDArray, list]],
        is_training: bool,
        image_mode: Optional[str] = "RGB",
    ):
        """
        Read documents, process them, and stack them. One sample has one document image.

        Parameters
        ----------
        document_features
            One sample has one document image column in a pd.DataFrame.
        feature_modalities
            What modality each column belongs to.
        is_training
            Whether to process document images in the training mode.
        image_mode
            A string which defines the type and depth of a pixel in the image.
            For example, RGB, RGBA, CMYK, and etc.

        Returns
        -------
        A dictionary containing one sample's document and its features.
        """
        ret = {}
        for per_col_name, per_col_image_features in document_features.items():
            try:
                # Process PDF documents.
                if feature_modalities[per_col_name] == DOCUMENT_PDF:
                    from pdf2image import convert_from_path

                    # Convert PDF to PIL images.
                    doc_images = convert_from_path(per_col_image_features[0])
                    if doc_images:
                        doc_image = doc_images[0].convert(image_mode)
                        words, normalized_word_boxes = self.get_ocr_features(per_col_image_features[0], doc_image)
                    else:
                        raise ValueError("Failed to convert PDF to images.")
                else:
                    # Process document image.
                    with PIL.Image.open(per_col_image_features[0]) as doc_image:
                        doc_image = doc_image.convert(image_mode)
                        words, normalized_word_boxes = self.get_ocr_features(per_col_image_features[0], doc_image)
            except ImportError as e:
                raise e
            except Exception as e:
                if self.missing_value_strategy.lower() == "zero":
                    logger.debug(f"Using a zero image due to '{e}'")
                    doc_image = PIL.Image.new(image_mode, (self.size, self.size), color=0)
                    doc_image = doc_image.convert(image_mode)
                    words = ""  # empty words
                    normalized_word_boxes = [self.pad_token_box]
                else:
                    raise e

            if is_training:
                doc_image = self.train_processor(doc_image)
            else:
                doc_image = self.val_processor(doc_image)

            if self.is_text_only_flag:
                # Padding of token_boxes up the bounding boxes to the sequence length.
                token_boxes = []
                all_tokens = []
                for word, box in zip(words, normalized_word_boxes):
                    # Tokenize bounding box separately.
                    word_tokens = self.tokenizer.tokenize(word)
                    all_tokens.append(word_tokens)
                    token_boxes.extend([box] * len(word_tokens))

                encoding = self.tokenizer(
                    " ".join(words), padding="max_length", truncation=True, return_token_type_ids=True
                )

                # Truncation of token_boxes
                special_tokens_count = 2  # one cls token and one sep token
                if len(token_boxes) > self.text_max_len - special_tokens_count:
                    token_boxes = token_boxes[: (self.text_max_len - special_tokens_count)]
                # add bounding boxes of cls + sep tokens
                token_boxes = [self.pad_token_box] + token_boxes + [self.sep_token_box]

                padding_length = self.text_max_len - sum(encoding.attention_mask)
                token_boxes += [self.pad_token_box] * padding_length

            else:
                if not isinstance(words, list):
                    words = [words]
                encoding = self.tokenizer(
                    words,
                    boxes=normalized_word_boxes,
                    padding="max_length",
                    truncation=True,
                    return_token_type_ids=True,
                )
                token_boxes = encoding.bbox

            ret.update(
                {
                    self.text_token_ids_key: np.array(encoding.input_ids, dtype=np.int32),
                    self.text_attention_mask_key: encoding.attention_mask,
                    self.text_bbox_key: np.array(token_boxes, dtype=np.int32),
                    self.text_segment_ids_key: np.array(encoding.token_type_ids, dtype=np.int32),
                }
            )
            if not self.is_text_only_flag:
                ret.update({self.document_pixel_value_key: doc_image})

        return ret

    def save_tokenizer(
        self,
        path: str,
    ):
        """
        Save the text tokenizer and record its relative paths, e.g, hf_text.

        Parameters
        ----------
        path
            The root path of saving.

        """
        save_path = os.path.join(path, self.prefix)
        self.tokenizer.save_pretrained(save_path)
        self.tokenizer = self.prefix

    def load_tokenizer(
        self,
        path: str,
    ):
        """
        Load saved text tokenizers. If text/ner processors already have tokenizers,
        then do nothing.

        Parameters
        ----------
        path
            The root path of loading.

        Returns
        -------
        A list of text/ner processors with tokenizers loaded.
        """
        if isinstance(self.tokenizer, str):
            load_path = os.path.join(path, self.tokenizer)
            self.tokenizer = get_pretrained_tokenizer(
                tokenizer_name=self.tokenizer_name,
                checkpoint_name=load_path,
            )

    def __call__(
        self,
        all_features: Dict[str, Union[NDArray, list]],
        feature_modalities: Dict[str, Union[NDArray, list]],
        is_training: bool,
    ) -> Dict:
        """
        Extract one sample's multimodal data.

        Parameters
        ----------
        all
            All the raw input data.
        is_training
            Whether to do processing in the training mode.

        Returns
        -------
        A dictionary containing one sample's features and/or labels.
        """

        ret = self.process_one_sample(all_features, feature_modalities, is_training)

        return ret



import logging
import os
from typing import Dict, List, Optional, Tuple, Union

logger = logging.getLogger(__name__)

# TODO: Try a better workaround to lazy import tensorrt package.
tensorrt_imported = False
if not tensorrt_imported:
    try:
        import tensorrt  # Unused but required by TensorrtExecutionProvider

        tensorrt_imported = True
    except:
        # We silently omit the import failure here to avoid overwhelming warning messages in case of multi-gpu.
        tensorrt_imported = False


def onnx_get_dynamic_axes(input_keys: List[str]):
    dynamic_axes = {}
    for k in input_keys:
        if "token_ids" in k or "segment_ids" in k:
            dynamic_axes[k] = {0: "batch_size", 1: "seq_length"}
        elif "valid_length" in k or k.startswith("numerical") or k.startswith("timm_image"):
            dynamic_axes[k] = {0: "batch_size"}

    return dynamic_axes


def get_provider_name(provider_config: Union[str, tuple]) -> str:
    if isinstance(provider_config, tuple):
        provider_name = provider_config[0]
    else:
        assert isinstance(provider_config, str), "input provider config is expected to be either str or tuple"
        provider_name = provider_config
    return provider_name


class OnnxModule(object):
    """
    OnnxModule is as a replacement of torch.nn.Module for running forward pass with onnxruntime.

    The module can be generated with MultiModalPredictor.export_tensorrt(),
    so that we can predict with TensorRT by simply replacing predictor._model with OnnxModule.
    """

    def __init__(self, onnx_path: Union[str, bytes], providers: Optional[Union[dict, List[str]]] = None):
        """
        Parameters
        ----------
        onnx_path : str or bytes
            The file path (or bytes) of the onnx model that need to be executed in onnxruntime.
        providers : dict or str, default=None
            A list of execution providers for model prediction in onnxruntime.
        """
        import onnx
        import onnxruntime as ort

        if isinstance(onnx_path, bytes):
            onnx_model = onnx.load_model_from_string(onnx_path)
        else:
            if not os.path.exists(onnx_path):
                raise FileNotFoundError(f"failed to located onnx file at {onnx_path}")

            logger.info("Loading ONNX file from path {}...".format(onnx_path))
            onnx_model = onnx.load(onnx_path)

        if providers is None:
            if isinstance(onnx_path, str):
                dirname = os.path.dirname(os.path.abspath(onnx_path))
                cache_path = os.path.join(dirname, "model_trt")
            else:
                cache_path = None
            providers = [
                (
                    "TensorrtExecutionProvider",
                    {
                        "device_id": 0,
                        "trt_max_workspace_size": 2147483648,
                        "trt_fp16_enable": True,
                        "trt_engine_cache_path": cache_path,
                        "trt_engine_cache_enable": True,
                    },
                ),
                (
                    "CUDAExecutionProvider",
                    {
                        "device_id": 0,
                        "arena_extend_strategy": "kNextPowerOfTwo",
                        "gpu_mem_limit": 2 * 1024 * 1024 * 1024,
                        "cudnn_conv_algo_search": "EXHAUSTIVE",
                        "do_copy_in_default_stream": True,
                    },
                ),
                ("CPUExecutionProvider", {}),
            ]

        if len(providers) == 1 and get_provider_name(providers[0]) == "TensorrtExecutionProvider":
            if not tensorrt_imported:
                raise ImportError(
                    "tensorrt package is not installed. The package can be install via `pip install tensorrt`."
                )

        self.sess = ort.InferenceSession(onnx_model.SerializeToString(), providers=providers)

        if get_provider_name(providers[0]) == "TensorrtExecutionProvider" and tensorrt_imported:
            assert "TensorrtExecutionProvider" in self.sess.get_providers(), (
                f"unexpected TensorRT compilation failure: TensorrtExecutionProvider not in providers ({self.sess.get_providers()}). "
                "Make sure onnxruntime package gets lazy imported everywhere."
            )

        inputs = self.sess.get_inputs()
        outputs = self.sess.get_outputs()
        self.input_names = [i.name for i in inputs]
        self.output_names = [i.name for i in outputs]

    def __call__(self, *args):
        """
        Make the module callable like torch.nn.Module, while running forward pass with onnxruntime.

        Parameters
        ----------
        args : list of torch.Tensor
            A list of torch.Tensor that are inputs of the model.

        Returns
        -------
        onnx_outputs : list of torch.Tensor
            A list of torch.Tensor that are outputs of the model.
        """
        import torch

        input_dict = {k: args[i].cpu().numpy() for i, k in enumerate(self.input_names)}
        onnx_outputs = self.sess.run(self.output_names, input_dict)
        onnx_outputs = onnx_outputs[:3]
        onnx_outputs = [torch.from_numpy(out) for out in onnx_outputs]
        return onnx_outputs

    def to(self, *args):
        """A dummy function that act as torch.nn.Module.to() function"""

        class DummyModel:
            def eval():
                pass

        return DummyModel



import json
import os
from enum import Enum

import pandas as pd


class TaskType(Enum):
    """
    Currently supported task type
    """

    IMAGE_CLASSIFICATION = "image_classification"
    OBJECT_DETECTION = "object_detection"
    CUSTOMIZE = "customize"
    TEXT_CLASSIFICATION = "text_classification"
    TEXT_SUMMARIZATION = "text_summarization"
    NAMED_ENTITY_RECOGNITION = "named_entity_recognition"


# preset data and label columns of some given label-studio template (except customized)
columns_template = {
    "image_classification": {
        "csv": {"data_columns": ["image"], "label_columns": ["choice"]},
        "json": {"data_columns": ["image"], "label_columns": ["value.choices"]},
        "json_min": {"data_columns": ["image"], "label_columns": ["choice"]},
    },
    "object_detection": {
        "csv": {"data_columns": ["image"], "label_columns": ["label"]},
        "json": {
            "data_columns": ["image"],
            "label_columns": [
                "original_width",
                "original_height",
                "value.x",
                "value.y",
                "value.width",
                "value.height",
                "value.rotation",
                "value.rectanglelabels",
            ],
        },
        "json_min": {"data_columns": ["image"], "label_columns": ["label"]},
    },
    "named_entity_recognition": {
        "csv": {"data_columns": ["text"], "label_columns": ["label"]},
        "json": {
            "data_columns": ["text"],
            "label_columns": ["value.start", "value.end", "value.text", "value.labels"],
        },
        "json_min": {"data_columns": ["text"], "label_columns": ["label"]},
    },
}


def read_from_labelstudio_csv(path, data_columns, label_columns):
    """
    read data from exported .csv files and return an Dataframe

    params:
    - path: str: the path of the exported file
    - data_columns: list[str]: the key/column names of the data
    - label_columns: list[str]: the key/column names of the label
    """
    if not os.path.exists(path):
        raise OSError("annotation file path not exists.")
    labelstudio_csv = pd.read_csv(path)
    df = pd.DataFrame()
    columns = labelstudio_csv.columns
    # extracting columns for given data and label column names
    for col in data_columns:
        if col in columns:
            df[col] = labelstudio_csv[col]
        else:
            print("skip '{}' for not in the export data columns".format(col))
    for col in label_columns:
        if col in columns:
            df[col] = labelstudio_csv[col]
        else:
            print("skip '{}' for not in the export label columns".format(col))
    return df


def read_from_labelstudio_json(path, data_columns, label_columns):
    """
    read data from exported .json files and return an Dataframe
    (include the JSON and JSON-MIN files from Label-Studio)

    params:
    - path: str: the path of the exported file
    - data_columns: list[str]: the key/column names of the data
    - label_columns: list[str]: the key/column names of the label
    """
    if not os.path.exists(path):
        raise OSError("annotation file path not exists.")

    with open(path, mode="r") as fp:
        label_studio_json = json.load(fp)
        if len(label_studio_json) == 0:
            raise ValueError("ERROR: empty export file")

        if "annotations" in label_studio_json[0]:
            # the file is export through JSON
            # JSON file parsing is available to specify nested data
            annotation = pd.json_normalize(
                data=label_studio_json, record_path=["annotations", "result"], meta=["data"]
            )
            data = pd.DataFrame(annotation["data"].values.tolist())

            # annotation: labeling content
            # data: data content/url/...
            df = pd.DataFrame()

            columns = data.columns
            for col in data_columns:
                if col in columns:
                    df[col] = data[col]
                else:
                    print("skip '{}' for not in the export data columns".format(col))

            columns = annotation.columns
            for col in label_columns:
                if col in columns:
                    df[col] = annotation[col]
                else:
                    print("skip '{}' for not in the export label columns".format(col))

            return df

        else:
            # the file is export through JSON-MIN

            df = pd.DataFrame()
            annotation_table = pd.json_normalize(data=label_studio_json)

            for col in data_columns:
                if col in annotation_table:
                    df[col] = annotation_table[col]
                else:
                    print("skip '{}' for not in the export data columns".format(col))

            for col in label_columns:
                if col in annotation_table:
                    df[col] = annotation_table[col]
                else:
                    print("skip '{}' for not in the export label columns".format(col))

            return df


def get_dataframes_by_path(path, data_columns, label_columns):
    """
    get the data frames by path and given data, label column names,
    and return the Dataframe

    params:
    - path: str: the path of the exported file
    - data_columns: list[str]: the key/column names of the data
    - label_columns: list[str]: the key/column names of the label
    """

    # get file extension through os
    _, file_extension = os.path.splitext(path)
    file_extension = file_extension[1:]

    if file_extension == "csv":
        return read_from_labelstudio_csv(path, data_columns, label_columns)
    elif file_extension == "json" or file_extension == "json_min":
        return read_from_labelstudio_json(path, data_columns, label_columns)
    else:
        raise OSError("current file extension {} is not supported.".format(file_extension))


class LabelStudioReader:
    """
    a tool that transfer label-studio export file to the dataframe for autogluon training. Docs available at
    https://github.com/autogluon/autogluon/tree/master/examples/automm/label_studio/label_studio_export_reader
    """

    def __init__(self, host=None):
        self.default_host = "http://localhost:8080" if not host else host
        print(
            "NOTE: the default label-studio host is {},if you want to get data from an running label-studio url, "
            "please set 'ls_host_on' to True".format(self.default_host)
        )
        self.templates = columns_template

    def set_labelstudio_host(self, host):
        if host:
            self.default_host = host
            print("set label-studio default host to {}".format(host))

    def get_columns_by_type(self, type, format):
        """
        for tasks that labeled from a given Label-Studio Template with the names
        of the template not modified, the data_columns and label_columns can be
        achieved through the preset template.
        params:
        - type: Enum of the template type(e.g: image classification).See class TaskType.
        - format: str: file export format: (csv, json, json_min)
        """
        if type.value != "customize":
            data_columns = self.templates[type.value][format]["data_columns"]
            label_columns = self.templates[type.value][format]["label_columns"]
            return data_columns, label_columns
        else:
            return [], []

    def process_data_str(self, s: str, ls_online):
        """
        used for lambda expression of the data str. See the detailed documentation
        params:
        - s: str, the origin data str
        - ls_online: boolean: whether the label-studio host should be provided
        """
        if ls_online:
            if s.startswith("/data/"):
                # label-studio imported data.
                # data are accessible when label-studio host is on
                # data can be addressed through a temporal ls-host-based URL
                return self.default_host + s
            else:
                # url, text content,...
                # need no further process
                return s

        else:
            local_storage_prefix = "/data/local-files/?d="
            upload_prefix = "/data/upload"
            if s.startswith(local_storage_prefix):
                # data imported from Label-Studio local storage
                # changed to the local path of the file
                return s[len(local_storage_prefix) :]
            elif s.startswith(upload_prefix):
                # the uploaded file can not be accessed
                print("Warning: cannot read {} with the label-studio host off.".format(s))
                return s
            # TODO: add s3, gcd, redis support
            else:
                return s

    def from_image_classification(self, path, ls_host_on, data_columns=None, label_columns=None):
        """
        data: image files
        process export file from label-studio image classification template,
        return the overall and label Dataframe for Autogluon input
        param:
        - path: str: the local path of exported file
        - ls_host_on: boolean: if the label-studio host is needed
        - data_columns: list[str](Optional) key or column names of data
        - label_columns: list[str](Optional) key or column names of label
        """

        if not os.path.exists(path):
            raise OSError("annotation file path not exists.")

        # get file extension with os, distinguish JSON and JSON-MIN
        _, file_extension = os.path.splitext(path)
        file_extension = file_extension[1:]
        if file_extension == "json":
            with open(path, mode="r") as fp:
                json_content = json.load(fp)
                if "annotations" not in json_content[0]:
                    file_extension = "json_min"

        # get the data and label columns through default image classification template
        default_data, default_label = self.get_columns_by_type(TaskType.IMAGE_CLASSIFICATION, format=file_extension)
        # use user preset if there's any
        data = data_columns if data_columns else default_data
        label = label_columns if label_columns else default_label

        df = get_dataframes_by_path(path, data, label)

        columns = df.columns
        for col in data:
            if col in columns:
                # process the data file content/url according to ls_host_on
                df[col] = df[col].apply(lambda s: self.process_data_str(s, ls_host_on))
            else:
                print("skip '{}' for not in the data column names.".format(col))

        return df, df[label]

    def from_named_entity_recognition(self, path, data_columns=None, label_columns=None):
        """
        data: text
        process export file from label-studio NER template,
        return the overall and label Dataframe for Autogluon input
        param:
        - path: str: the local path of exported file
        - data_columns: list[str](Optional) key or column names of data
        - label_columns: list[str](Optional) key or column names of label
        """

        if not os.path.exists(path):
            raise OSError("annotation file path not exists.")

        # get file extension with os,distinguish JSON and JSON-MIN
        _, file_extension = os.path.splitext(path)
        file_extension = file_extension[1:]
        if file_extension == "json":
            with open(path, mode="r") as fp:
                json_content = json.load(fp)
                if "annotations" not in json_content[0]:
                    file_extension = "json_min"

        # get the data and label columns through default NER template
        default_data, default_label = self.get_columns_by_type(
            TaskType.NAMED_ENTITY_RECOGNITION, format=file_extension
        )
        data = data_columns if data_columns else default_data
        label = label_columns if label_columns else default_label

        df = get_dataframes_by_path(path, data, label)
        return df, df[label]

    def from_customize(self, path, ls_host_on, data_columns, label_columns):
        """
        process export file from customized template,
        return the overall and label Dataframe for Autogluon input
        param:
        - path: str: the local path of exported file
        - ls_host_on: boolean: if the label-studio host is needed
        - data_columns: list[str](Optional) key or column names of data
        - label_columns: list[str](Optional) key or column names of label
        """

        if not os.path.exists(path):
            raise OSError("annotation file path not exists.")

        _, file_extension = os.path.splitext(path)
        file_extension = file_extension[1:]
        if file_extension == "json":
            with open(path, mode="r") as fp:
                json_content = json.load(fp)
                if "annotations" not in json_content[0]:
                    file_extension = "json_min"

        data, label = data_columns, label_columns

        df = get_dataframes_by_path(path, data, label)

        columns = df.columns
        for col in data:
            if col in columns:
                df[col] = df[col].apply(lambda s: self.process_data_str(s, ls_host_on))
            else:
                print("skip '{}' for not in the data column names.".format(col))

        return df, df[label]



import logging
from typing import Any, Dict, List, Optional, Tuple, Union

import torch
from torch import nn, optim
from transformers import Adafactor
from transformers.trainer_pt_utils import get_parameter_names

from ..constants import (
    BIT_FIT,
    COLUMN_FEATURES,
    CONV_LORA,
    FEATURES,
    IA3,
    IA3_BIAS,
    IA3_LORA,
    IA3_LORA_BIAS,
    IA3_LORA_NORM,
    IA3_NORM,
    LORA,
    LORA_BIAS,
    LORA_NORM,
    NORM_FIT,
    PEFT_STRATEGIES,
)

logger = logging.getLogger(__name__)


def get_optimizer(
    optim_type: str,
    optimizer_grouped_parameters,
    lr: float,
    weight_decay: float,
    eps: Optional[float] = 1e-6,
    betas: Optional[Tuple[float, float]] = (0.9, 0.999),
    momentum: Optional[float] = 0.9,
):
    """
    Choose a Pytorch optimizer based on its name.

    Parameters
    ----------
    optim_type
        Name of optimizer.
    optimizer_grouped_parameters
        The model parameters to be optimized.
    lr
        Learning rate.
    weight_decay
        Optimizer weight decay.
    eps
        Optimizer eps.
    betas
        Optimizer betas.
    momentum
        Momentum used in the SGD optimizer.

    Returns
    -------
    A Pytorch optimizer.
    """
    if optim_type == "adamw":
        optimizer = optim.AdamW(
            optimizer_grouped_parameters,
            lr=lr,
            weight_decay=weight_decay,
            eps=eps,
            betas=betas,
        )
    elif optim_type == "adam":
        optimizer = optim.Adam(
            optimizer_grouped_parameters,
            lr=lr,
            weight_decay=weight_decay,
        )
    elif optim_type == "sgd":
        optimizer = optim.SGD(
            optimizer_grouped_parameters,
            lr=lr,
            weight_decay=weight_decay,
            momentum=momentum,
        )
    elif optim_type == "adafactor":
        optimizer = Adafactor(
            optimizer_grouped_parameters,
            lr=lr,
            weight_decay=weight_decay,
            scale_parameter=True,  # Generally recommended to enable scaling
            relative_step=False,
            warmup_init=False,
        )
    else:
        raise ValueError(f"unknown optimizer: {optim_type}")

    return optimizer


def get_weight_decay_param_names(model: nn.Module):
    """
    Set the layer normalization parameters and other layers' bias parameters not to use weight decay.

    Parameters
    ----------
    model
        A Pytorch model.

    Returns
    -------
    A list of parameter names not using weight decay.
    """
    # By default, we should not apply weight decay for all the norm layers
    decay_param_names = get_parameter_names(
        model,
        [nn.LayerNorm, nn.BatchNorm1d, nn.BatchNorm2d, nn.BatchNorm3d, nn.GroupNorm],
    )
    decay_param_names = [
        name
        for name in decay_param_names
        if (
            "bias" not in name
            and "cls_token" not in name
            and "categorical_feature_tokenizer" not in name
            and "numerical_feature_tokenizer" not in name
        )
    ]
    return decay_param_names


def get_norm_layer_param_names(model: nn.Module):
    """
    Get parameters associated with the normalization layers

    Parameters
    ----------
    model
        A Pytorch model

    Returns
    -------
    norm_param_names
        A list of normalization parameter names
    """
    all_param_names = [name for name, _ in model.named_parameters()]
    all_param_names_except_norm_names = get_parameter_names(
        model,
        [nn.LayerNorm, nn.BatchNorm1d, nn.BatchNorm2d, nn.BatchNorm3d, nn.GroupNorm],
    )
    norm_param_names = [name for name in all_param_names if name not in all_param_names_except_norm_names]
    return norm_param_names


def get_peft_param_names(norm_param_names: List[str], peft: Optional[str] = None, extra_params: Optional[List] = None):
    """
     Get the list of trainable parameters according to the provided efficient finetuning method.

    Parameters
    ----------
    norm_param_names
        The parameters associated with the normalization layers
    peft
        Efficient finetuning strategy. Trainable parameters will be adjusted according to the method.
    extra_params
        Extra parameters to train.

    Returns
    -------
    Get list of trainable parameter names according to the provided efficient finetuning method.
    """
    peft_param_names = []

    if peft == BIT_FIT:
        peft_param_names.append(".*bias*.")
    elif peft == NORM_FIT:
        peft_param_names.append(".*bias*.")
        peft_param_names += norm_param_names
    elif peft in [LORA, IA3, IA3_LORA, CONV_LORA]:
        peft_param_names.append(".*lora_*.")
    elif peft in [LORA_BIAS, IA3_BIAS, IA3_LORA_BIAS]:
        peft_param_names.append(".*lora_*.")
        peft_param_names.append(".*bias*.")
    elif peft in [LORA_NORM, IA3_NORM, IA3_LORA_NORM]:
        peft_param_names.append(".*lora_*.")
        peft_param_names.append(".*bias*.")
        peft_param_names += norm_param_names
    elif peft is not None:
        raise NotImplementedError(
            f"The efficient finetuning strategy '{peft}'"
            f" is not supported. We only support"
            f" {', '.join(PEFT_STRATEGIES)}."
        )

    if extra_params:
        peft_param_names.extend(extra_params)

    return peft_param_names


def remove_parameters_without_grad(
    grouped_parameters: List[Dict],
):
    """
    Remove layers

    Parameters
    ----------
    grouped_parameters
        The grouped parameters or their names output from lr_choice.

    Returns
    -------
    The updated grouped parameters or their names.
    """
    for group_idx, group_param in enumerate(grouped_parameters):
        updated_params = []
        for p in group_param["params"]:
            if p.requires_grad:
                updated_params.append(p)
        grouped_parameters[group_idx]["params"] = updated_params

    return grouped_parameters


def gather_column_features(
    output: Dict[str, Dict],
    column_names: Union[str, List[str]],
):
    """
    Gather column features from models' outputs.
    For each feature name in one model's output, we enumerate the provided column names to see
    whether (partial) the provided columns share one cls feature or they have independent features.

    TODO: return features' masks and use them to filter the losses.

    Parameters
    ----------
    output
        The models' outputs.
    column_names
        The columns whose features we want to get.

    Returns
    -------
    The gathered feature vectors. Each sample should only have one feature vector.
    """
    if isinstance(column_names, str):
        column_names = [column_names]

    gathered_features = []
    # logger.debug(f"gather features for columns: {column_names}")
    for per_model_name, per_model_output in output.items():
        # logger.debug(f"gather column features from model: {per_model_name}")
        for feature_name in per_model_output[COLUMN_FEATURES][FEATURES]:
            # logger.debug(f"processing feature: {feature_name}")
            columns_share_one_feature = []
            for col_name in column_names:
                if col_name in feature_name:
                    # this column feature is part of the cls feature
                    if not (feature_name.startswith(col_name) and feature_name.endswith(col_name)):
                        columns_share_one_feature.append(col_name)
                        # logger.debug(f"column {col_name} is included in feature {feature_name}")
                    else:  # this column's feature is independent of other columns'
                        gathered_features.append(per_model_output[COLUMN_FEATURES][FEATURES][col_name])
                        # logger.debug(f"col_name {col_name} has an independent feature in model: {per_model_name}")

            # two or more columns share one cls feature, and no other columns share it.
            if len(columns_share_one_feature) > 0:
                assert (
                    len("_".join(columns_share_one_feature)) == len(feature_name)
                ), f"model `{per_model_name}`'s cls feature name `{feature_name}` doesn't match `{columns_share_one_feature}`"
                gathered_features.append(per_model_output[COLUMN_FEATURES][FEATURES][feature_name])

    if len(gathered_features) > 1:
        # currently only support features of the same shape
        assert all(
            per_features.shape == gathered_features[0].shape for per_features in gathered_features
        ), "Currently we only support gathering features of the same dimension."

    if len(gathered_features) == 0:
        raise ValueError(f"No features are found for columns names {column_names}.")

    gathered_features = torch.stack(gathered_features, dim=0).mean(dim=0)  # (b, d)

    return gathered_features



from .lit_distiller import DistillerLitModule
from .lit_matcher import MatcherLitModule
from .lit_mmdet import MMDetLitModule
from .lit_module import LitModule
from .lit_ner import NerLitModule
from .lit_semantic_seg import SemanticSegmentationLitModule
from .losses import get_aug_loss_func, get_loss_func, get_matcher_loss_func, get_matcher_miner_func
from .metrics import (
    CustomHitRate,
    compute_ranking_score,
    compute_score,
    get_minmax_mode,
    get_stopping_threshold,
    get_torchmetric,
    infer_metrics,
)
from .utils import get_norm_layer_param_names, get_peft_param_names



from .coverage_metrics import Coverage
from .hit_rate_metrics import CustomHitRate
from .ranking_metrics import compute_ranking_score
from .utils import get_torchmetric, compute_score, get_minmax_mode, get_stopping_threshold, infer_metrics



import functools
import logging
import warnings
from typing import Dict, List, Optional, Tuple, Union

import evaluate
import torchmetrics
from torch.nn import functional as F
from torchmetrics.detection.mean_ap import MeanAveragePrecision

from autogluon.core.metrics import Scorer, compute_metric, get_metric

from ...constants import (
    ACC,
    ACCURACY,
    AVERAGE_PRECISION,
    BER,
    COSINE_EMBEDDING_LOSS,
    COVERAGE,
    CROSS_ENTROPY,
    DETECTION_METRICS,
    DIRECT_LOSS,
    EM,
    F1,
    F1_MACRO,
    F1_MICRO,
    F1_WEIGHTED,
    FM,
    HIT_RATE,
    IOU,
    LOG_LOSS,
    MAE,
    MATCHING_METRICS,
    MATCHING_METRICS_WITHOUT_PROBLEM_TYPE,
    MAX,
    METRIC_MODE_MAP,
    MIN,
    MULTICLASS,
    NER_TOKEN_F1,
    OVERALL_ACCURACY,
    OVERALL_F1,
    PEARSONR,
    QUADRATIC_KAPPA,
    R2,
    RECALL,
    RETRIEVAL_METRICS,
    RMSE,
    ROC_AUC,
    ROOT_MEAN_SQUARED_ERROR,
    SM,
    SPEARMANR,
    Y_PRED,
    Y_PRED_PROB,
    Y_TRUE,
)
from .coverage_metrics import Coverage
from .hit_rate_metrics import CustomHitRate
from .semantic_seg_metrics import COD_METRICS_NAMES, Balanced_Error_Rate, Binary_IoU, Multiclass_IoU

logger = logging.getLogger(__name__)


def compute_score(
    metric_data: dict,
    metric: Union[str, Scorer],
    pos_label: Optional[int] = 1,
) -> float:
    """
    Use sklearn to compute the score of one metric.

    Parameters
    ----------
    metric_data
        A dictionary with the groundtruth (Y_TRUE) and predicted values (Y_PRED, Y_PRED_PROB).
        The predicted class probabilities are required to compute the roc_auc score.
    metric
        The name of metric or the function of metric to compute.
    pos_label
        The encoded label (0 or 1) of binary classification's positive class.

    Returns
    -------
    Computed score.
    """
    if isinstance(metric, str) and metric in [OVERALL_ACCURACY, OVERALL_F1]:
        metric = evaluate.load("seqeval")
        warnings.filterwarnings("ignore")
        for p in metric_data[Y_TRUE]:
            if "_" in p:
                print(p)
        for p in metric_data[Y_PRED]:
            if "_" in p:
                print(p)
        return metric.compute(references=metric_data[Y_TRUE], predictions=metric_data[Y_PRED])

    metric = get_metric(metric)

    y = metric_data[Y_TRUE]
    if metric.needs_proba or metric.needs_threshold:
        y_pred_proba = metric_data[Y_PRED_PROB]
        y_pred_proba = (
            y_pred_proba if y_pred_proba.shape[1] > 2 else y_pred_proba[:, pos_label]
        )  # only use pos_label for binary classification
        return metric.convert_score_to_original(
            compute_metric(y=y, y_pred_proba=y_pred_proba, metric=metric, weights=None)
        )
    else:
        y_pred = metric_data[Y_PRED]

        # TODO: This is a hack. Doesn't support `f1_macro`, `f1_micro`, `f1_weighted`, or custom `f1` metrics with different names.
        # TODO: Longterm the solution should be to have the input data to this function use the internal representation without the original class names. This way `pos_label` would not need to be specified.
        if metric.name == F1:  # only for binary classification
            y = (y == pos_label).astype(int)
            y_pred = (y_pred == pos_label).astype(int)

        return metric.convert_score_to_original(compute_metric(y=y, y_pred=y_pred, metric=metric, weights=None))


def infer_metrics(
    problem_type: Optional[str] = None,
    eval_metric: Optional[Union[str, Scorer]] = None,
    validation_metric_name: Optional[str] = None,
    is_matching: Optional[bool] = False,
):
    """
    Infer the validation metric and the evaluation metric if not provided.
    Validation metric is for early-stopping and selecting the best model checkpoints.
    Evaluation metric is to report performance to users.

    Parameters
    ----------
    problem_type
        Type of problem.
    eval_metric
        Name of evaluation metric provided by users.
    validation_metric_name
        The provided validation metric name
    is_matching
        Whether is matching.

    Returns
    -------
    validation_metric_name
        Name of validation metric.
    eval_metric_name
        Name of evaluation metric.
    """
    is_customized = False
    if eval_metric is None:
        eval_metric_name = None
    elif isinstance(eval_metric, str):
        eval_metric_name = eval_metric
    elif isinstance(eval_metric, Scorer):
        eval_metric_name = eval_metric.name
        is_customized = True
    else:
        raise TypeError(f"eval_metric can be a str, a Scorer, or None, but is type: {type(eval_metric)}")

    if problem_type is not None:
        from ...utils.problem_types import PROBLEM_TYPES_REG

        problem_property = PROBLEM_TYPES_REG.get(problem_type)

    if is_matching:
        if eval_metric_name is not None:
            # if eval_metric_name is a valid metric
            if eval_metric_name.lower() in METRIC_MODE_MAP.keys():
                validation_metric_name = eval_metric_name
                return validation_metric_name, eval_metric_name
            elif eval_metric_name.lower() in RETRIEVAL_METRICS:
                # Currently only support recall as validation metric in retrieval tasks.
                validation_metric_name = RECALL
                return validation_metric_name, eval_metric_name

        # When eval_metric_name is either None or not supported:
        # Fallback based on problem type unless it's a customized metric
        if problem_type is None:
            validation_metric_name, fallback_evaluation_metric = MATCHING_METRICS_WITHOUT_PROBLEM_TYPE
        elif problem_type in MATCHING_METRICS:
            validation_metric_name, fallback_evaluation_metric = MATCHING_METRICS[problem_type]
        else:
            raise NotImplementedError(f"Problem type: {problem_type} is not yet supported for matching!")
        if not is_customized:
            if eval_metric_name is not None:
                warnings.warn(
                    f"Metric {eval_metric_name} is not supported as the evaluation metric for {problem_type} in matching tasks."
                    f"The evaluation metric is changed to {fallback_evaluation_metric} by default."
                )
            eval_metric_name = fallback_evaluation_metric
        return validation_metric_name, eval_metric_name

    if eval_metric_name is not None:
        # Infer evaluation metric
        if eval_metric_name.lower() not in problem_property.supported_evaluation_metrics and not is_customized:
            warnings.warn(
                f"Metric {eval_metric_name} is not supported as the evaluation metric for {problem_type}. "
                f"The evaluation metric is changed to {problem_property.fallback_evaluation_metric} by default."
            )
            if problem_property.fallback_evaluation_metric is not None:
                eval_metric_name = problem_property.fallback_evaluation_metric
            else:
                # Problem types like extract_embedding does not need a eval/val metric
                return None, None

        # Infer validation metric
        if eval_metric_name.lower() in problem_property.supported_validation_metrics:
            validation_metric_name = eval_metric_name
        else:
            if problem_property.fallback_validation_metric is not None:
                validation_metric_name = problem_property.fallback_validation_metric
    else:
        eval_metric_name = problem_property.fallback_evaluation_metric
        validation_metric_name = problem_property.fallback_validation_metric

    return validation_metric_name, eval_metric_name


def get_minmax_mode(
    metric_name: Union[str, Scorer],
):
    """
    Get minmax mode based on metric name

    Parameters
    ----------
    metric_name
        A string representing metric

    Returns
    -------
    mode
        The min/max mode used in selecting model checkpoints.
        - min
             Its means that smaller metric is better.
        - max
            It means that larger metric is better.
    """
    if isinstance(metric_name, str):
        assert (
            metric_name in METRIC_MODE_MAP
        ), f"{metric_name} is not a supported metric. Options are: {METRIC_MODE_MAP.keys()}"
        return METRIC_MODE_MAP.get(metric_name)
    else:
        return MAX if metric_name.greater_is_better else MIN


def get_stopping_threshold(metric_name: str):
    """
    Get the metric threshold for early stopping.

    Parameters
    ----------
    metric_name
        Name of validation metric.

    Returns
    -------
    The stopping threshold.
    """
    try:
        metric = get_metric(metric_name)
        stopping_threshold = metric.optimum - metric._sign * 1e-7
    except:
        stopping_threshold = None

    return stopping_threshold


def get_torchmetric(
    metric_name: str,
    num_classes: Optional[int] = None,
    is_matching: Optional[bool] = False,
    problem_type: Optional[str] = None,
):
    """
    Obtain a torchmerics.Metric from its name.
    Define a customized metric function in case that torchmetrics doesn't support some metric.

    Parameters
    ----------
    metric_name
        Name of metric.
    num_classes
        Number of classes.
    is_matching
        Whether is matching.
    problem_type
        Type of problem, e.g., binary and multiclass.

    Returns
    -------
    torchmetrics.Metric
        A torchmetrics.Metric object.
    custom_metric_func
        A customized metric function.
    """
    metric_name = metric_name.lower()
    if metric_name in [ACC, ACCURACY, OVERALL_ACCURACY]:
        # use MULTICLASS since the head output dim is 2 for the binary problem type.
        return torchmetrics.Accuracy(task=MULTICLASS, num_classes=num_classes), None
    elif metric_name == NER_TOKEN_F1:
        return torchmetrics.F1Score(task=MULTICLASS, num_classes=num_classes, ignore_index=1), None
    elif metric_name in [RMSE, ROOT_MEAN_SQUARED_ERROR]:
        return torchmetrics.MeanSquaredError(squared=False), None
    elif metric_name == R2:
        return torchmetrics.R2Score(), None
    elif metric_name == QUADRATIC_KAPPA:
        return (
            torchmetrics.CohenKappa(task=problem_type, num_classes=num_classes, weights="quadratic"),
            None,
        )
    elif metric_name == ROC_AUC:
        return torchmetrics.AUROC(task=problem_type, num_classes=num_classes), None
    elif metric_name == AVERAGE_PRECISION:
        return torchmetrics.AveragePrecision(task=problem_type, num_classes=num_classes)
    elif metric_name in [LOG_LOSS, CROSS_ENTROPY]:
        return torchmetrics.MeanMetric(), functools.partial(F.cross_entropy, reduction="none")
    elif metric_name == COSINE_EMBEDDING_LOSS:
        return torchmetrics.MeanMetric(), functools.partial(F.cosine_embedding_loss, reduction="none")
    elif metric_name == PEARSONR:
        return torchmetrics.PearsonCorrCoef(), None
    elif metric_name == SPEARMANR:
        if is_matching:  # TODO: add support for matching.
            raise ValueError("spearman relation is not supported for matching yet.")
        else:
            return torchmetrics.SpearmanCorrCoef(), None
    elif metric_name == F1:
        return torchmetrics.F1Score(task=problem_type, num_classes=num_classes), None
    elif metric_name in [F1_MACRO, F1_MICRO, F1_WEIGHTED]:
        average = metric_name.split("_")[1]
        return torchmetrics.F1Score(task=problem_type, num_classes=num_classes, average=average), None
    elif metric_name in DETECTION_METRICS:
        return (
            MeanAveragePrecision(box_format="xyxy", iou_type="bbox", class_metrics=False),
            None,
        )  # TODO: remove parameter hardcodings here, and add class_metrics
    elif metric_name == DIRECT_LOSS:
        return (
            torchmetrics.MeanMetric(nan_strategy="warn"),
            None,
        )  # This only works for detection where custom_metric is not required for BaseAggregator
    elif metric_name in [RECALL, HIT_RATE]:
        if is_matching:
            return CustomHitRate(), None
        else:  # TODO: support recall for general classification tasks.
            raise ValueError("Recall is not supported yet.")
    elif metric_name == BER:
        return Balanced_Error_Rate(), None
    elif metric_name in [SM, EM, FM, MAE]:
        return COD_METRICS_NAMES[metric_name], None
    elif metric_name == IOU:
        if num_classes == 1:
            return Binary_IoU(), None
        else:
            return Multiclass_IoU(num_classes=num_classes), None
    elif metric_name == COVERAGE:
        return Coverage(), None
    else:
        raise ValueError(f"Unknown metric {metric_name}")



import json
import logging
import os
from datetime import timedelta
from typing import Dict, List, Optional, Union

import pandas as pd
from omegaconf import DictConfig
from torch import nn

from ..constants import BBOX, DDP, MAP, MULTI_IMAGE_MIX_DATASET, OBJECT_DETECTION, XYWH
from ..data import (
    BaseDataModule,
    MultiImageMixDataset,
    MultiModalFeaturePreprocessor,
    infer_rois_column_type,
    split_train_tuning_data,
)
from ..models import create_fusion_model
from ..optim import MMDetLitModule
from ..utils import (
    check_if_packages_installed,
    cocoeval,
    convert_pred_to_xywh,
    convert_result_df,
    extract_from_output,
    from_coco_or_voc,
    get_detection_classes,
    object_detection_data_to_df,
    save_result_coco_format,
    setup_save_path,
)
from .base import BaseLearner

logger = logging.getLogger(__name__)


class ObjectDetectionLearner(BaseLearner):
    def __init__(
        self,
        label: Optional[str] = None,  # TODO: can we let users customize label?
        problem_type: Optional[str] = OBJECT_DETECTION,
        presets: Optional[str] = None,
        eval_metric: Optional[str] = None,
        hyperparameters: Optional[dict] = None,
        path: Optional[str] = None,
        verbosity: Optional[int] = 2,
        num_classes: Optional[int] = None,
        classes: Optional[list] = None,
        category_ids: Optional[list] = None,
        warn_if_exist: Optional[bool] = True,
        enable_progress_bar: Optional[bool] = None,
        pretrained: Optional[bool] = True,
        validation_metric: Optional[str] = None,
        sample_data_path: Optional[str] = None,  # TODO: can we use train/predict data instead?
        **kwargs,
    ):
        """
        Parameters
        ----------
        num_classes
            Number of classes. Used in classification.
            If this is specified and is different from the pretrained model's output,
            the model's head will be changed to have <num_classes> output.
        classes
            All classes in this dataset.
        sample_data_path
            This is used for automatically inference num_classes, classes, or label.
        """
        super().__init__(
            problem_type=problem_type,
            presets=presets,
            eval_metric=eval_metric,
            hyperparameters=hyperparameters,
            path=path,
            verbosity=verbosity,
            warn_if_exist=warn_if_exist,
            enable_progress_bar=enable_progress_bar,
            pretrained=pretrained,
            validation_metric=validation_metric,
        )
        check_if_packages_installed(problem_type=self._problem_type)

        self._config = self.get_config_per_run(config=self._config, hyperparameters=hyperparameters)

        self._output_shape = num_classes
        self._classes = classes
        self._category_ids = category_ids
        self._sample_data_path = sample_data_path

        # TODO: merge object detection and open vocabulary object detection
        self._label_column = "label"
        if self._sample_data_path is not None:
            self._classes, self._category_ids = get_detection_classes(self._sample_data_path)
            self._output_shape = len(self._classes)

        # TODO: merge _detection_anno_train and detection_anno_train?
        self._detection_anno_train = None
        self.detection_anno_train = None

        self._log_filters += [
            ".*Creating a tensor from a list of numpy.ndarrays is extremely slow..*",
        ]

    @property
    def classes(self):
        """
        Return the classes of object detection.
        """
        if self._model.model.CLASSES is not None and self._classes is not None:
            assert self._classes == self._model.model.CLASSES, f"{self._classes}\n{self._model.model.CLASSES}"
        return self._classes if self._classes is not None else self._model.model.CLASSES

    @property
    def category_ids(self):
        """
        Return the classes of object detection.
        """
        return self._category_ids

    def setup_detection_train_tuning_data(self, max_num_tuning_data, seed, train_data, tuning_data):
        if isinstance(train_data, str):
            self._detection_anno_train = train_data
            train_data = from_coco_or_voc(
                train_data,
                "train",
                coco_root=self._config.model.mmdet_image.coco_root,
            )  # TODO: Refactor to use convert_data_to_df
            if tuning_data is not None:
                self.detection_anno_train = tuning_data
                tuning_data = from_coco_or_voc(
                    tuning_data,
                    "val",
                    coco_root=self._config.model.mmdet_image.coco_root,
                )  # TODO: Refactor to use convert_data_to_df
                if max_num_tuning_data is not None:
                    if len(tuning_data) > max_num_tuning_data:
                        tuning_data = tuning_data.sample(
                            n=max_num_tuning_data, replace=False, random_state=seed
                        ).reset_index(drop=True)
        elif isinstance(train_data, pd.DataFrame):
            self._detection_anno_train = None
            # sanity check dataframe columns
            train_data = object_detection_data_to_df(
                train_data,
                coco_root=self._config.model.mmdet_image.coco_root,
            )
            if tuning_data is not None:
                self.detection_anno_train = tuning_data
                tuning_data = object_detection_data_to_df(
                    tuning_data,
                    coco_root=self._config.model.mmdet_image.coco_root,
                )
                if max_num_tuning_data is not None:
                    if len(tuning_data) > max_num_tuning_data:
                        tuning_data = tuning_data.sample(
                            n=max_num_tuning_data, replace=False, random_state=seed
                        ).reset_index(drop=True)
        else:
            raise TypeError(f"Expected train_data to have type str or pd.DataFrame, but got type: {type(train_data)}")
        return train_data, tuning_data

    def prepare_train_tuning_data(
        self,
        train_data: Union[pd.DataFrame, str],
        tuning_data: Optional[Union[pd.DataFrame, str]],
        holdout_frac: Optional[float],
        max_num_tuning_data: Optional[int],
        seed: Optional[int],
    ):
        # TODO: remove self from calling setup_detection_train_tuning_data()
        train_data, tuning_data = self.setup_detection_train_tuning_data(
            train_data=train_data,
            tuning_data=tuning_data,
            max_num_tuning_data=max_num_tuning_data,
            seed=seed,
        )

        if tuning_data is None:
            train_data, tuning_data = split_train_tuning_data(
                data=train_data,
                holdout_frac=holdout_frac,
                problem_type=self._problem_type,
                label_column=self._label_column,
                random_state=seed,
            )

        self._train_data = train_data
        self._tuning_data = tuning_data

    def infer_output_shape(self, **kwargs):
        # TODO: support inferring output during fit()?
        assert self._output_shape is not None, f"output_shape should have been set in the learner initialization."

    def init_pretrained(self):
        super().init_pretrained()
        self._model = create_fusion_model(
            config=self._config,
            pretrained=self._pretrained,
            num_classes=self._output_shape,
            classes=self._classes,
        )

    def fit(
        self,
        train_data: Union[pd.DataFrame, str],
        presets: Optional[str] = None,
        tuning_data: Optional[Union[pd.DataFrame, str]] = None,
        max_num_tuning_data: Optional[int] = None,
        time_limit: Optional[int] = None,
        save_path: Optional[str] = None,
        hyperparameters: Optional[Union[str, Dict, List[str]]] = None,
        column_types: Optional[Dict] = None,
        holdout_frac: Optional[float] = None,
        seed: Optional[int] = 0,
        standalone: Optional[bool] = True,
        hyperparameter_tune_kwargs: Optional[Dict] = None,
        clean_ckpts: Optional[bool] = True,
        **kwargs,
    ):
        training_start = self.on_fit_start(presets=presets)
        self.setup_save_path(save_path=save_path)
        self.prepare_train_tuning_data(
            train_data=train_data,
            tuning_data=tuning_data,
            holdout_frac=holdout_frac,
            max_num_tuning_data=max_num_tuning_data,
            seed=seed,
        )
        self.infer_column_types(column_types=column_types)
        self.infer_validation_metric()
        self.update_hyperparameters(
            hyperparameters=hyperparameters,
            hyperparameter_tune_kwargs=hyperparameter_tune_kwargs,
        )
        self.fit_sanity_check()
        self.prepare_fit_args(
            time_limit=time_limit,
            seed=seed,
            standalone=standalone,
            clean_ckpts=clean_ckpts,
        )
        fit_returns = self.execute_fit()
        self.on_fit_end(
            training_start=training_start,
            strategy=fit_returns.get("strategy", None),
            strict_loading=fit_returns.get("strict_loading", True),
            standalone=standalone,
            clean_ckpts=clean_ckpts,
        )

        return self

    def get_datamodule_per_run(
        self,
        df_preprocessor,
        data_processors,
        per_gpu_batch_size,
        num_workers,
        model_config=None,
        predict_data=None,
        is_train=True,
    ):
        datamodule_kwargs = dict(
            df_preprocessor=df_preprocessor,
            data_processors=data_processors,
            per_gpu_batch_size=per_gpu_batch_size,
            num_workers=num_workers,
        )
        if is_train:
            val_use_training_mode = (self._problem_type == OBJECT_DETECTION) and (self._validation_metric_name != MAP)
            datamodule_kwargs.update(
                dict(validate_data=self._tuning_data, val_use_training_mode=val_use_training_mode)
            )
            if (
                self._problem_type == OBJECT_DETECTION
                and model_config is not None
                and MULTI_IMAGE_MIX_DATASET in model_config
            ):
                train_dataset = MultiImageMixDataset(
                    data=self._train_data,
                    preprocessor=[df_preprocessor],
                    processors=[data_processors],
                    model_config=model_config,
                    id_mappings=None,
                    is_training=True,
                )
                datamodule_kwargs.update(dict(train_dataset=train_dataset))
            else:
                datamodule_kwargs.update(dict(train_data=self._train_data))
        else:
            datamodule_kwargs.update(dict(predict_data=predict_data))

        datamodule = BaseDataModule(**datamodule_kwargs)
        return datamodule

    def get_strategy_per_run(self, num_gpus, config):
        if num_gpus <= 1:
            strategy = "auto"
        else:
            strategy = DDP

        return strategy

    def update_num_gpus_by_strategy(self, strategy, num_gpus):
        if strategy == DDP and self._fit_called:
            num_gpus = 1  # While using DDP, we can only use single gpu after fit is called

        return num_gpus

    def get_optim_kwargs_per_run(self, config, validation_metric, custom_metric_func):
        return dict(
            optim_type=config.optim.optim_type,
            lr_choice=config.optim.lr_choice,
            lr_schedule=config.optim.lr_schedule,
            lr=config.optim.lr,
            lr_decay=config.optim.lr_decay,
            end_lr=config.optim.end_lr,
            lr_mult=config.optim.lr_mult,
            weight_decay=config.optim.weight_decay,
            warmup_steps=config.optim.warmup_steps,
            track_grad_norm=config.optim.track_grad_norm,
            validation_metric=validation_metric,
            validation_metric_name=self._validation_metric_name,
            custom_metric_func=custom_metric_func,
        )

    def get_litmodule_per_run(
        self,
        model: Optional[nn.Module] = None,
        optim_kwargs: Optional[dict] = None,
        is_train=True,
    ):
        if self._problem_type == OBJECT_DETECTION:
            LightningModule = MMDetLitModule
        else:
            raise TypeError(f"problem type {self._problem_type} is not supported by ObjectDetectionLearner.")

        if is_train:
            return LightningModule(
                model=model,
                **optim_kwargs,
            )
        else:
            return LightningModule(model=self._model)

    def get_model_per_run(self, model, config):
        if model is None:
            model = create_fusion_model(
                config=config,
                num_classes=self._output_shape,
                classes=self._classes,
            )
        return model

    def fit_per_run(
        self,
        max_time: timedelta,
        save_path: str,
        ckpt_path: str,
        resume: bool,
        enable_progress_bar: bool,
        seed: int,
        hyperparameters: Optional[Union[str, Dict, List[str]]] = None,
        advanced_hyperparameters: Optional[Dict] = None,
        config: Optional[Dict] = None,
        df_preprocessor: Optional[MultiModalFeaturePreprocessor] = None,
        data_processors: Optional[Dict] = None,
        model: Optional[nn.Module] = None,
        standalone: bool = True,
        clean_ckpts: bool = True,
    ):
        self.on_fit_per_run_start(seed=seed, save_path=save_path)
        config = self.get_config_per_run(config=config, hyperparameters=hyperparameters)
        df_preprocessor = self.get_df_preprocessor_per_run(
            df_preprocessor=df_preprocessor,
            config=config,
        )
        config = self.update_config_by_data_per_run(config=config, df_preprocessor=df_preprocessor)
        model = self.get_model_per_run(model=model, config=config)
        model = self.compile_model_per_run(config=config, model=model)
        data_processors = self.get_data_processors_per_run(
            data_processors=data_processors,
            config=config,
            model=model,
            advanced_hyperparameters=advanced_hyperparameters,
        )
        validation_metric, custom_metric_func = self.get_validation_metric_per_run()
        if max_time == timedelta(seconds=0):
            return dict(
                config=config,
                df_preprocessor=df_preprocessor,
                data_processors=data_processors,
                model=model,
            )
        datamodule = self.get_datamodule_per_run(
            df_preprocessor=df_preprocessor,
            data_processors=data_processors,
            per_gpu_batch_size=config.env.per_gpu_batch_size,
            num_workers=config.env.num_workers,
            model_config=model.config,
        )
        optim_kwargs = self.get_optim_kwargs_per_run(
            config=config,
            validation_metric=validation_metric,
            custom_metric_func=custom_metric_func,
        )
        litmodule = self.get_litmodule_per_run(
            model=model,
            optim_kwargs=optim_kwargs,
        )
        callbacks = self.get_callbacks_per_run(save_path=save_path, config=config, litmodule=litmodule)
        plugins = self.get_plugins_per_run(model=model)
        tb_logger = self.get_tb_logger(save_path=save_path)
        num_gpus, strategy = self.get_num_gpus_and_strategy_per_run(config=config)
        precision = self.get_precision_per_run(num_gpus=num_gpus, precision=config.env.precision)
        grad_steps = self.get_grad_steps(num_gpus=num_gpus, config=config)
        strategy = self.get_strategy_per_run(num_gpus=num_gpus, config=config)
        config = self.post_update_config_per_run(
            config=config,
            num_gpus=num_gpus,
            precision=precision,
            strategy=strategy,
        )
        trainer = self.init_trainer_per_run(
            num_gpus=num_gpus,
            config=config,
            precision=precision,
            strategy=strategy,
            max_time=max_time,
            callbacks=callbacks,
            tb_logger=tb_logger,
            grad_steps=grad_steps,
            plugins=plugins,
            enable_progress_bar=enable_progress_bar,
        )

        self.run_trainer(
            trainer=trainer,
            litmodule=litmodule,
            datamodule=datamodule,
            ckpt_path=ckpt_path,
            resume=resume,
        )
        self.on_fit_per_run_end(
            save_path=save_path,
            standalone=standalone,
            trainer=trainer,
            config=config,
            df_preprocessor=df_preprocessor,
            data_processors=data_processors,
            model=model,
        )

        return dict(
            config=config,
            df_preprocessor=df_preprocessor,
            data_processors=data_processors,
            model=model,
            best_score=trainer.callback_metrics[f"val_{self._validation_metric_name}"].item(),
            strategy=strategy,
        )

    def predict_per_run(
        self,
        data: Union[pd.DataFrame, dict, list],
        realtime: Optional[bool],
        requires_label: bool,
        barebones: Optional[bool] = False,
    ) -> List[Dict]:
        """
        Perform inference for learner.

        Parameters
        ----------
        data
            The data for inference.
        realtime
            Whether use realtime inference.
        requires_label
            Whether uses label during inference.
        barebones
            Whether to run in “barebones mode”, where all lightning's features that may impact raw speed are disabled.

        Returns
        -------
        A list of output dicts.
        """
        data = self.on_predict_per_run_start(data=data)
        column_types = self.infer_column_types(
            column_types=self._column_types,
            data=data,
            is_train=False,
        )
        column_types = infer_rois_column_type(
            column_types=column_types,
            data=data,
        )
        df_preprocessor = self.get_df_preprocessor_per_run(
            df_preprocessor=self._df_preprocessor,
            data=data,
            column_types=column_types,
            is_train=False,
        )
        if self._fit_called:
            df_preprocessor._column_types = self.update_image_column_types(data=data)
        data_processors = self.get_data_processors_per_run(
            data_processors=self._data_processors,
            requires_label=requires_label,
            is_train=False,
        )
        num_gpus, strategy = self.get_num_gpus_and_strategy_per_run(
            predict_data=data,
            is_train=False,
        )
        precision = self.get_precision_per_run(
            num_gpus=num_gpus,
            precision=self._config.env.precision,
            cpu_only_warning=False,
        )
        batch_size = self.get_predict_batch_size_per_run(num_gpus=num_gpus, strategy=strategy)
        realtime, num_gpus, barebones = self.update_realtime_for_interactive_env(
            realtime=realtime,
            num_gpus=num_gpus,
            barebones=barebones,
            strategy=strategy,
        )
        datamodule = self.get_datamodule_per_run(
            df_preprocessor=df_preprocessor,
            data_processors=data_processors,
            per_gpu_batch_size=batch_size,
            num_workers=self._config.env.num_workers_inference,
            predict_data=data,
            is_train=False,
        )
        pred_writer = self.get_pred_writer(strategy=strategy)
        callbacks = self.get_callbacks_per_run(pred_writer=pred_writer, is_train=False)
        litmodule = self.get_litmodule_per_run(is_train=False)
        trainer = self.init_trainer_per_run(
            num_gpus=num_gpus,
            precision=precision,
            strategy=strategy,
            callbacks=callbacks,
            barebones=barebones,
            is_train=False,
        )
        outputs = self.run_trainer(
            trainer=trainer,
            litmodule=litmodule,
            datamodule=datamodule,
            pred_writer=pred_writer,
            is_train=False,
        )
        outputs = self.collect_predictions(
            outputs=outputs,
            trainer=trainer,
            pred_writer=pred_writer,
            num_gpus=num_gpus,
        )
        self.on_predict_per_run_end(trainer=trainer)

        # TODO: remove this by adjusting the return format of mmdet_image or lit_mmdet.
        if pred_writer is None and self._problem_type == OBJECT_DETECTION:
            outputs = [output for batch_outputs in outputs for output in batch_outputs]

        return outputs

    def evaluate_coco(
        self,
        anno_file_or_df: str,
        metrics: str,
        return_pred: Optional[bool] = False,
        eval_tool: Optional[str] = None,
    ):
        """
        Evaluate object detection model on a test dataset in COCO format.

        Parameters
        ----------
        anno_file_or_df
            The annotation file in COCO format
        metrics
            Metrics used for evaluation.
        return_pred
            Whether to return the prediction result of each row.
        eval_tool
            The eval_tool for object detection. Could be "pycocotools" or "torchmetrics".
        """
        if isinstance(anno_file_or_df, str):
            anno_file = anno_file_or_df
            data = from_coco_or_voc(
                anno_file,
                "test",
                coco_root=self._config.model.mmdet_image.coco_root,
            )  # TODO: maybe remove default splits hardcoding (only used in VOC)
            if os.path.isdir(anno_file):
                eval_tool = "torchmetrics"  # we can only use torchmetrics for VOC format evaluation.
        else:
            # during validation, it will call evaluate with df as input
            anno_file = self._detection_anno_train
            data = anno_file_or_df

        outputs = self.predict_per_run(
            data=data,
            realtime=False,
            requires_label=True,
        )  # outputs shape: num_batch, 1(["bbox"]), batch_size, 2(if using mask_rcnn)/na, 80, n, 5

        # Cache prediction results as COCO format # TODO: refactor this
        self._save_path = setup_save_path(
            old_save_path=self._save_path,
            warn_if_exist=False,
        )
        cocoeval_cache_path = os.path.join(self._save_path, "object_detection_result_cache.json")
        eval_results = cocoeval(
            outputs=outputs,
            data=data,
            anno_file=anno_file,
            cache_path=cocoeval_cache_path,
            metrics=metrics,
            tool=eval_tool,
        )
        if return_pred:
            return eval_results, outputs
        else:
            return eval_results

    def evaluate(
        self,
        data: Union[pd.DataFrame, dict, list, str],
        metrics: Optional[Union[str, List[str]]] = None,
        return_pred: Optional[bool] = False,
        realtime: Optional[bool] = False,
        eval_tool: Optional[str] = None,
        **kwargs,
    ):
        """
        Evaluate model on a test dataset.

        Parameters
        ----------
        data
            A dataframe, containing the same columns as the training data.
            Or a str, that is a path of the annotation file for detection.
        metrics
            A list of metric names to report.
            If None, we only return the score for the stored `_eval_metric_name`.
        return_pred
            Whether to return the prediction result of each row.
        realtime
            Whether to do realtime inference, which is efficient for small data (default False).
            If provided None, we would infer it on based on the data modalities
            and sample number.
        eval_tool
            The eval_tool for object detection. Could be "pycocotools" or "torchmetrics".

        Returns
        -------
        A dictionary with the metric names and their corresponding scores.
        Optionally return a dataframe of prediction results.
        """
        self.ensure_predict_ready()

        if realtime:
            return NotImplementedError(f"Current problem type {self._problem_type} does not support realtime predict.")
        if isinstance(data, str):
            return self.evaluate_coco(
                anno_file_or_df=data,
                metrics=metrics,
                return_pred=return_pred,
                eval_tool=eval_tool,
            )
        else:
            data = object_detection_data_to_df(
                data,
                coco_root=self._config.model.mmdet_image.coco_root,
            )
            return self.evaluate_coco(
                anno_file_or_df=data,
                metrics=metrics,
                return_pred=return_pred,
                eval_tool="torchmetrics",
            )

    def predict(
        self,
        data: Union[pd.DataFrame, dict, list, str],
        as_pandas: Optional[bool] = None,
        as_coco: Optional[bool] = True,
        realtime: Optional[bool] = False,
        save_results: Optional[bool] = None,
        **kwargs,
    ):
        """
        Predict values for the label column of new data.

        Parameters
        ----------
        data
            The data to make predictions for. Should contain same column names as training data and
            follow same format (except for the `label` column).
        as_pandas
            Whether to return the output as a pandas DataFrame(Series) (True) or Instance Data (False).
            For the definition of Instance Data in MMDetection/MMEngine, see
            https://github.com/open-mmlab/mmengine/blob/698782f9203a6bfcc0e445047fd2300796ecbf0f/mmengine/structures/instance_data.py#L34
        as_coco
            Whether to save the output as a COCO json file (True) or pandas DataFrame (False).
        realtime
            Whether to do realtime inference, which is efficient for small data (default False).
            If provided None, we would infer it on based on the data modalities
            and sample number.
        save_results
            Whether to save the prediction results (only works for detection now)

        Returns
        -------
        Array of predictions, one corresponding to each row in given dataset.
        """
        self.ensure_predict_ready()

        if as_pandas is None:
            as_pandas = True  # return pandas dataframe by default

        ret_type = BBOX

        # only supports coco/voc format for OBJECT_DETECTION
        if self._problem_type == OBJECT_DETECTION:
            data_path = data
            data_df = object_detection_data_to_df(
                data_path,
                coco_root=self._config.model.mmdet_image.coco_root,
            )
            if self._label_column not in data_df:
                self._label_column = None

        outputs = self.predict_per_run(
            data=data_df,
            realtime=realtime,
            requires_label=False,
        )
        pred = extract_from_output(outputs=outputs, ret_type=ret_type)

        self._save_path = setup_save_path(
            old_save_path=self._save_path,
            warn_if_exist=False,
        )
        result_path = os.path.join(self._save_path, "result.txt")

        pred_df = convert_result_df(
            pred=convert_pred_to_xywh(pred) if self._model.output_bbox_format == XYWH else pred,
            data=data_df,
            detection_classes=self.classes,
            result_path=result_path,
        )

        if save_results:
            self._save_path = setup_save_path(
                old_save_path=self._save_path,
                warn_if_exist=False,
            )
            if as_coco:
                result_path = os.path.join(self._save_path, "result.json")
                save_result_coco_format(
                    data_path=data_path,
                    pred=pred,
                    category_ids=self.category_ids,
                    result_path=result_path,
                    coco_root=self._config.model.mmdet_image.coco_root,
                )
            else:
                pred_df.to_csv(result_path, index=False)
            logger.info(f"Saved detection results {'as coco' if as_coco else 'as dataframe'} to {result_path}")

        if as_pandas:
            return pred_df
        else:
            if self._model.output_bbox_format == XYWH:
                pred = convert_pred_to_xywh(pred)
            return pred

    def predict_proba(
        self,
        data: Union[pd.DataFrame, dict, list],
        as_pandas: Optional[bool] = None,
        as_multiclass: Optional[bool] = True,
        realtime: Optional[bool] = False,
        **kwargs,
    ):
        raise NotImplementedError("Object detection doesn't support calling `predict_proba` yet.")

    def extract_embedding(
        self,
        data: Union[pd.DataFrame, dict, list],
        as_tensor: Optional[bool] = False,
        as_pandas: Optional[bool] = False,
        realtime: Optional[bool] = False,
        **kwargs,
    ):
        raise NotImplementedError("Object detection doesn't support calling `extract_embedding` yet.")

    @staticmethod
    def _load_metadata(
        learner,
        path: str,
        resume: Optional[bool] = False,
        verbosity: Optional[int] = 3,
    ):
        learner = super()._load_metadata(learner=learner, path=path, resume=resume, verbosity=verbosity)
        learner._data_processors = None
        return learner

    def save(
        self,
        path: str,
        standalone: Optional[bool] = True,
        config: Optional[DictConfig] = None,
        model: Optional[nn.Module] = None,
        df_preprocessor: Optional[MultiModalFeaturePreprocessor] = None,
        data_processors: Optional[Dict] = None,
        fit_called: Optional[bool] = None,
        save_model: Optional[bool] = True,
    ):
        super().save(
            path=path,
            standalone=standalone,
            config=config,
            model=model,
            df_preprocessor=df_preprocessor,
            data_processors=data_processors,
            fit_called=fit_called,
            save_model=save_model,
        )
        assets_path = os.path.join(path, "assets.json")
        with open(assets_path, "r") as fp:
            assets = json.load(fp)
            assets.update(
                {
                    "classes": self._classes,
                }
            )
        os.remove(assets_path)
        with open(assets_path, "w") as fp:
            json.dump(assets, fp, ensure_ascii=True)



_base_ = ["../coco_detection.py", "../default_runtime.py"]
model = dict(
    type="DINO",
    num_queries=900,  # num_matching_queries
    with_box_refine=True,
    as_two_stage=True,
    data_preprocessor=dict(
        type="DetDataPreprocessor",
        mean=[123.675, 116.28, 103.53],
        std=[58.395, 57.12, 57.375],
        bgr_to_rgb=True,
        pad_size_divisor=1,
    ),
    backbone=dict(
        type="ResNet",
        depth=50,
        num_stages=4,
        out_indices=(1, 2, 3),
        frozen_stages=1,
        norm_cfg=dict(type="BN", requires_grad=False),
        norm_eval=True,
        style="pytorch",
        init_cfg=dict(type="Pretrained", checkpoint="torchvision://resnet50"),
    ),
    neck=dict(
        type="ChannelMapper",
        in_channels=[512, 1024, 2048],
        kernel_size=1,
        out_channels=256,
        act_cfg=None,
        norm_cfg=dict(type="GN", num_groups=32),
        num_outs=4,
    ),
    encoder=dict(
        num_layers=6,
        layer_cfg=dict(
            self_attn_cfg=dict(embed_dims=256, num_levels=4, dropout=0.0),  # 0.1 for DeformDETR
            ffn_cfg=dict(embed_dims=256, feedforward_channels=2048, ffn_drop=0.0),  # 1024 for DeformDETR
        ),
    ),  # 0.1 for DeformDETR
    decoder=dict(
        num_layers=6,
        return_intermediate=True,
        layer_cfg=dict(
            self_attn_cfg=dict(embed_dims=256, num_heads=8, dropout=0.0),  # 0.1 for DeformDETR
            cross_attn_cfg=dict(embed_dims=256, num_levels=4, dropout=0.0),  # 0.1 for DeformDETR
            ffn_cfg=dict(embed_dims=256, feedforward_channels=2048, ffn_drop=0.0),  # 1024 for DeformDETR
        ),  # 0.1 for DeformDETR
        post_norm_cfg=None,
    ),
    positional_encoding=dict(
        num_feats=128,
        normalize=True,
        offset=0.0,
        temperature=20,  # -0.5 for DeformDETR
    ),  # 10000 for DeformDETR
    bbox_head=dict(
        type="DINOHead",
        num_classes=80,
        sync_cls_avg_factor=True,
        loss_cls=dict(type="FocalLoss", use_sigmoid=True, gamma=2.0, alpha=0.25, loss_weight=1.0),  # 2.0 in DeformDETR
        loss_bbox=dict(type="L1Loss", loss_weight=5.0),
        loss_iou=dict(type="GIoULoss", loss_weight=2.0),
    ),
    dn_cfg=dict(  # TODO: Move to model.train_cfg ?
        label_noise_scale=0.5,
        box_noise_scale=1.0,  # 0.4 for DN-DETR
        group_cfg=dict(dynamic=True, num_groups=None, num_dn_queries=100),
    ),  # TODO: half num_dn_queries
    # training and testing settings
    train_cfg=dict(
        assigner=dict(
            type="HungarianAssigner",
            match_costs=[
                dict(type="FocalLossCost", weight=2.0),
                dict(type="BBoxL1Cost", weight=5.0, box_format="xywh"),
                dict(type="IoUCost", iou_mode="giou", weight=2.0),
            ],
        )
    ),
    test_cfg=dict(max_per_img=300),
)  # 100 for DeformDETR

# train_pipeline, NOTE the img_scale and the Pad's size_divisor is different
# from the default setting in mmdet.
train_pipeline = [
    dict(type="LoadImageFromFile", backend_args={{_base_.backend_args}}),  # nosec
    dict(type="LoadAnnotations", with_bbox=True),
    dict(type="RandomFlip", prob=0.5),
    dict(
        type="RandomChoice",
        transforms=[
            [
                dict(
                    type="RandomChoiceResize",
                    scales=[
                        (480, 1333),
                        (512, 1333),
                        (544, 1333),
                        (576, 1333),
                        (608, 1333),
                        (640, 1333),
                        (672, 1333),
                        (704, 1333),
                        (736, 1333),
                        (768, 1333),
                        (800, 1333),
                    ],
                    keep_ratio=True,
                )
            ],
            [
                dict(
                    type="RandomChoiceResize",
                    # The radio of all image in train dataset < 7
                    # follow the original implement
                    scales=[(400, 4200), (500, 4200), (600, 4200)],
                    keep_ratio=True,
                ),
                dict(type="RandomCrop", crop_type="absolute_range", crop_size=(384, 600), allow_negative_crop=True),
                dict(
                    type="RandomChoiceResize",
                    scales=[
                        (480, 1333),
                        (512, 1333),
                        (544, 1333),
                        (576, 1333),
                        (608, 1333),
                        (640, 1333),
                        (672, 1333),
                        (704, 1333),
                        (736, 1333),
                        (768, 1333),
                        (800, 1333),
                    ],
                    keep_ratio=True,
                ),
            ],
        ],
    ),
    dict(type="PackDetInputs"),
]
train_dataloader = dict(dataset=dict(filter_cfg=dict(filter_empty_gt=False), pipeline=train_pipeline))

# optimizer
optim_wrapper = dict(
    type="OptimWrapper",
    optimizer=dict(type="AdamW", lr=0.0001, weight_decay=0.0001),  # 0.0002 for DeformDETR
    clip_grad=dict(max_norm=0.1, norm_type=2),
    paramwise_cfg=dict(custom_keys={"backbone": dict(lr_mult=0.1)}),
)  # custom_keys contains sampling_offsets and reference_points in DeformDETR  # noqa

# learning policy
max_epochs = 12
train_cfg = dict(type="EpochBasedTrainLoop", max_epochs=max_epochs, val_interval=1)

val_cfg = dict(type="ValLoop")
test_cfg = dict(type="TestLoop")

param_scheduler = [dict(type="MultiStepLR", begin=0, end=max_epochs, by_epoch=True, milestones=[11], gamma=0.1)]

# NOTE: `auto_scale_lr` is for automatically scaling LR,
# USER SHOULD NOT CHANGE ITS VALUES.
# base_batch_size = (8 GPUs) x (2 samples per GPU)
auto_scale_lr = dict(base_batch_size=16)
